import numpy
import zarr

from ..constants.image import MD_SIZE_S, MD_SIZE_C, MD_SIZE_Z, MD_SIZE_T, MD_SIZE_Y, MD_SIZE_X, MD_SERIES_NAME
from ..constants.reader import ZARR_FILETYPE

from ..reader import Reader

import os

import re
import logging
from lxml import etree

LOGGER = logging.getLogger(__name__)

FORMAT_TESTER = re.compile(r"\.zarr([\\/]|$)", flags=re.IGNORECASE)
SUPPORTED_SCHEMES = {'file', 's3'}

class NGFFReader(Reader):
    """
    A reader for OME-NGFF files with the .zarr extension. Supports both 'normal' and
    'HCS' storage modes.

    OME-NGFF files below spec version 0.4 which are stored on S3/object storage
    should have the zarr.convenience.consolidate_metadata() function run on them
    to provide optimal performance over a network.

    Other non-OME .zarr format images may be readable, but are not strictly supported.
    """

    reader_name = "OME-NGFF"
    variable_revision_number = 1
    supported_filetypes = {'.zarr', '.ome.zarr'}
    supported_schemes = SUPPORTED_SCHEMES

    # Reader cache maps a path to a tuple of (zarr_root_group, series_map).
    ZARR_READER_CACHE = {}

    def __init__(self, image_file):
        super().__init__(image_file)

        self._reader = None
        # List of tuples denoting (array_path, array_name)
        self._series_map = None
        self.root, self.group = ZARR_FILETYPE.split(self.file.url, maxsplit=1)
        self.planes_extracted = False

    def __del__(self):
        self.close()

    def get_reader(self):
        if self._reader is not None:
            return self._reader
        elif self.root in NGFFReader.ZARR_READER_CACHE:
            self._reader, self._series_map = NGFFReader.ZARR_READER_CACHE[self.root]
        else:
            store = zarr.storage.FSStore(self.root)
            if self.root.lower().startswith('s3'):
                LOGGER.info("Zarr is stored on S3, will try to read directly.")
                if '.zmetadata' in store:
                    # Zarr has consolidated metadata.
                    self._reader = zarr.convenience.open_consolidated(store, mode='r')
                else:
                    LOGGER.warning(f"Image is on S3 but lacks consolidated metadata. "
                                    f"This may degrade reading performance. URL: {self.root}")
                    self._reader = zarr.open(store, mode='r')
            elif not os.path.isdir(store.path):
                raise IOError("The file, \"%s\", does not exist." % self.root)
            else:
                self._reader = zarr.open(store, mode='r')
            NGFFReader.ZARR_READER_CACHE[self.root] = self._reader, None
        if self.group:
            LOGGER.warning(f"Reader had a group? {self.group}")
            return self._reader[self.group]
        return self._reader

    @classmethod
    def clear_cached_readers(cls):
        NGFFReader.ZARR_READER_CACHE.clear()

    def read(self,
             wants_metadata_rescale=False,
             series=None,
             index=None,
             c=None,
             z=None,
             t=None,
             xywh=None,
             channel_names=None,
             ):
        """Read a single plane from the image file.
        :param wants_metadata_rescale: if `True`, return a tuple of image and a
               tuple of (min, max) for range values of image dtype gathered from
               file metadata; if `False`, returns only the image
        :param c: read from this channel. `None` = read color image if multichannel
            or interleaved RGB.
        :param z: z-stack index
        :param t: time index
        :param series: series for ``.flex`` and similar multi-stack formats
        :param index: if `None`, fall back to ``zct``, otherwise load the indexed frame
        :param xywh: a (x, y, w, h) tuple
        :param channel_names: provide the channel names for the OME metadata
        """
        LOGGER.debug(f"Reading {c=}, {z=}, {t=}, {series=}, {index=}, {xywh=}")
        c2 = None if c is None else c + 1
        z2 = None if z is None else z + 1
        t2 = None if t is None else t + 1
        if xywh is not None:
            x, y, w, h = xywh
            x = round(x)
            y = round(y)
            x2 = x + w
            y2 = y + h
        else:
            y, y2, x, x2 = None, None, None, None
        reader = self.get_reader()
        if series is None:
            series = 0
        if self._series_map is None:
            arrays = self.find_arrays(reader, multiscales=False, first_only=not self.planes_extracted)
            self.build_series_map(arrays)
        if not self._series_map:
            raise IOError("Zarr appears to contain no valid series. Unable to load data.")
        series_path, series_name = self._series_map[series]
        series_reader = reader[series_path]
        num_dimensions = len(series_reader.shape)
        # Todo: Handle V4 Axis object in .zattrs once it releases
        if num_dimensions == 5:
            image = series_reader[t:t2, c:c2, z:z2, y:y2, x:x2]
        elif num_dimensions == 4:
            image = series_reader[c:c2, z:z2, y:y2, x:x2]
        elif num_dimensions == 3:
            image = series_reader[z:z2, y:y2, x:x2]
        elif num_dimensions == 2:
            image = series_reader[y:y2, x:x2]
        else:
            raise NotImplementedError(f"Unsupported dimensionality: {num_dimensions}")
        # Remove redundant axes
        image = numpy.squeeze(image)
        # C needs to be the last axis, but z should be first. Thank you CellProfiler.
        if len(image.shape) > 2 and z is not None:
            image = numpy.moveaxis(image, 0, -1)
        elif len(image.shape) > 3:
            image = numpy.moveaxis(image, 0, -1)

        if wants_metadata_rescale:
            scale = self.get_max_sample_value(reader)
            return image, scale
        return image

    def read_volume(self,
                    wants_metadata_rescale=False,
                    series=None,
                    c=None,
                    z=None,
                    t=None,
                    xywh=None,
                    channel_names=None,
                    ):
        return self.read(
            wants_metadata_rescale=wants_metadata_rescale,
            series=series,
            c=c,
            z=None,
            t=t,
            xywh=xywh,
            channel_names=channel_names
        )

    @classmethod
    def supports_format(cls, image_file, allow_open=False, volume=False):
        """This function needs to evaluate whether a given ImageFile object
        can be read by this reader class.

        Return value should be an integer representing suitability:
        -1 - 'I can't read this at all'
        1 - 'I am the one true reader for this format, don't even bother checking any others'
        2 - 'I am well-suited to this format'
        3 - 'I can read this format, but I might not be the best',
        4 - 'I can give it a go, if you must'

        The allow_open parameter dictates whether the reader is permitted to read the file when
        making this decision. If False the decision should be made using file extension only.
        Any opened files should be closed before returning.

        The volume parameter specifies whether the reader will need to return a 3D array.
        ."""
        if FORMAT_TESTER.search(image_file.url) is not None:
            head, tail = os.path.splitext(image_file.path)
            if tail.lower() not in ('', '.zarr'):
                # An especially mean user may be pointing at a non-zarr file within the zarr?
                return 2
            return 1
        return -1

    def close(self):
        # If your reader opens a file, this needs to release any active lock,
        self._reader = None

    def build_series_map(self, arrays):
        if self.root in NGFFReader.ZARR_READER_CACHE:
            _, series_map = NGFFReader.ZARR_READER_CACHE[self.root]
            if series_map is not None:
                self._series_map = series_map
                return
        self._series_map = [(array.path, name) for array, name in arrays]
        if self.planes_extracted:
            NGFFReader.ZARR_READER_CACHE[self.root] = self._reader, self._series_map

    def get_series_metadata(self):
        """Should return a dictionary with the following keys:
        Key names are in cellprofiler_core.constants.image
        MD_SIZE_S - int reflecting the number of series
        MD_SIZE_X - list of X dimension sizes, one element per series.
        MD_SIZE_Y - list of Y dimension sizes, one element per series.
        MD_SIZE_Z - list of Z dimension sizes, one element per series.
        MD_SIZE_C - list of C dimension sizes, one element per series.
        MD_SIZE_T - list of T dimension sizes, one element per series.
        MD_SERIES_NAME - list of series names, one element per series.
        """
        self.planes_extracted = True
        meta_dict = {
            MD_SIZE_S: 0,
            MD_SIZE_X: [],
            MD_SIZE_Y: [],
            MD_SIZE_Z: [],
            MD_SIZE_C: [],
            MD_SIZE_T: [],
            MD_SERIES_NAME: [],
        }
        reader = self.get_reader()
        arrays = self.find_arrays(reader, multiscales=False, first_only=False)
        self.build_series_map(arrays)
        series_count = len(arrays)
        meta_dict[MD_SIZE_S] = series_count
        for idx, (array, series_name) in enumerate(arrays):
            dims = list(array.shape)
            meta_dict[MD_SIZE_X].append(dims.pop())
            meta_dict[MD_SIZE_Y].append(dims.pop())
            meta_dict[MD_SIZE_Z].append(dims.pop() if dims else 1)
            meta_dict[MD_SIZE_C].append(dims.pop() if dims else 1)
            meta_dict[MD_SIZE_T].append(dims.pop() if dims else 1)
            meta_dict[MD_SERIES_NAME].append(series_name)
        return meta_dict

    @staticmethod
    def get_max_sample_value(zarr_obj):
        xml_path = (zarr_obj.store.path + '/OME/METADATA.ome.xml')
        if not os.path.exists(xml_path):
            return None

        root = etree.parse(xml_path)

        namespaces = {
            "ome": "http://www.openmicroscopy.org/Schemas/OME/2016-06"
        }

        # Find the MaxSampleValue field
        # only present in transitional bioformats2raw.layout
        # https://ngff.openmicroscopy.org/0.5/index.html#bf2raw
        max_sample_value = root.xpath(
            "//ome:XMLAnnotation[ome:Value/ome:OriginalMetadata/ome:Key='MaxSampleValue']"
            "/ome:Value/ome:OriginalMetadata/ome:Value",
            namespaces=namespaces
        )

        # Retrieve the text content of MaxSampleValue, if found
        if max_sample_value:
            try:
                max_sample_value = int(max_sample_value[0].text)
            except:
                max_sample_value = None
            return max_sample_value
        else:
            return None

    @staticmethod
    def iterate_groups(zarr_obj):
        if 'plate' in zarr_obj.attrs and 'wells' in zarr_obj.attrs['plate']:
            # This is an HCS zarr with plate metadata. Use that to find the series.
            num_fields = zarr_obj.attrs['plate']['field_count']
            column_names = zarr_obj.attrs['plate']['columns']
            row_names = zarr_obj.attrs['plate']['rows']
            for attribs in zarr_obj.attrs['plate']['wells']:
                for field_idx in range(num_fields):
                    yield [zarr_obj[f"{attribs['path']}/{field_idx}"],
                           f"r{row_names[attribs['row_index']]['name']}"
                           f"c{column_names[attribs['column_index']]['name']}"
                           f"f{field_idx}"]
        elif 'XXXOME' in zarr_obj:
            # This is a zarr with OME metadata stored as a group. Grab a series list from this.
            ome_group = zarr_obj['OME']
            if 'series' in ome_group.attrs:
                series_paths = ome_group.attrs['series']
                for idx, series_path in enumerate(series_paths):
                    target = zarr_obj[series_path]
                    yield [target, ""]
        elif 'multiscales' in zarr_obj.attrs:
            # User has pointed directly to a zarr series, we can assume no child groups.
            yield [zarr_obj, ""]
        else:
            # This is a simple or older zarr. Look for an ome.xml metadata file and read that.
            xml_path = (zarr_obj.store.path + '/OME/METADATA.ome.xml')
            if os.path.exists(xml_path):
                ns = '{http://www.openmicroscopy.org/Schemas/OME/2016-06}'
                tree = etree.parse(xml_path)
                root = tree.getroot()
                if ns not in root.tag:
                    raise Exception(f"Expected {ns} in XML file")
                for image in root.findall(f"{ns}Image"):
                    series_path = image.attrib["ID"].split(":")[1]
                    series_name = image.attrib.get("Name", series_path)
                    if series_path in zarr_obj:
                        yield [zarr_obj[series_path], series_name]
                    else:
                        LOGGER.warning(f"Series specified in OME-XML "
                                       f"was absent from the "
                                       f"zarr - {series_path}")
            else:
                # No XML or zattrs = some other software made this zarr.
                # We can try scanning for groups, but this is slow.
                # Scans will return nothing if the storage backend is object-based (e.g. S3)
                LOGGER.error("Unable to find built-in or XML metadata, "
                             "will try a file scan")
                for key in zarr_obj.group_keys():
                    yield [zarr_obj[key], key]

    def find_arrays(self, zarr_obj, multiscales=False, first_only=False):
        """
        Searches a zarr object and produces a list of arrays within that object.
        :param zarr_obj: zarr.Array or zarr.Group to scan.
        :param multiscales: True = Consider all resolutions as arrays. False = Only return resolution 0.
        :param first_only: True = only grab the first array, don't bother scanning the whole file.
        :return: List of tuples in format (zarr_array, name). Empty string denotes no special name.
        """
        if isinstance(zarr_obj, zarr.Array):
            # We're already pointed directly at an array object. No scan needed.
            return [(zarr_obj, "")]
        if self._series_map is not None:
            return [(zarr_obj[path], name) for path, name in self._series_map]
        # Groups should be a list of groups which contain arrays as children.
        result = []

        for group_obj, group_name in self.iterate_groups(zarr_obj):
            if isinstance(group_obj, zarr.Array):
                result.append((group_obj, group_name))
            elif 'multiscales' in group_obj.attrs:
                # The main reason we do this - do we want just resolution 0 or all resolutions?
                scales = group_obj.attrs['multiscales']
                if multiscales:
                    for params in scales[0]['datasets']:
                        key = params['path']
                        result.append((group_obj[key], group_name))
                else:
                    key = scales[0]['datasets'][0]['path']
                    result.append((group_obj[key], group_name))
            else:
                LOGGER.error(f"Subgroup has no contents? {group_obj.path}")
            if first_only:
                break
        return result




from ._integer import Integer
from ...._validation_error import ValidationError


class OddInteger(Integer):
    def test_valid(self, pipeline):
        super(self.__class__, self).test_valid(pipeline)

        value = self.str_to_value(self.value_text)

        if value % 2 == 0:
            raise ValidationError("Must be odd, was even", self)



from ._compound_filter_predicate import CompoundFilterPredicate
from ._directory_predicate import DirectoryPredicate
from ._does_not_predicate import DoesNotPredicate
from ._does_predicate import DoesPredicate
from ._extension_predicate import ExtensionPredicate
from ._file_predicate import FilePredicate
from ._file_predicate import FilterPredicate
from ._filter import Filter
from ._image_predicate import ImagePredicate
from ._metadata_predicate import MetadataPredicate
from ._regexp_filter_predicate import RegexpFilterPredicate



from ._filter_predicate import FilterPredicate


class DoesPredicate(FilterPredicate):
    """Pass the arguments through (no-op)"""

    SYMBOL = "does"

    def __init__(
        self,
        subpredicates,
        text="Does",
        doc="The rule passes if the condition to the right holds",
    ):
        super(self.__class__, self).__init__(
            self.SYMBOL, text, lambda x, f, *l: f(x, *l), subpredicates, doc=doc
        )



from ._run_exception import RunException


class PrepareRunException(RunException):
    """An event indicating an uncaught exception during the prepare_run phase"""

    def event_type(self):
        return "Prepare run exception"



from ._alphanumeric import Alphanumeric
from ._character import Character
from .name import CropImageName
from .name import ExternalImageName
from .name import FileImageName
from .name import GridName
from .name import ImageName
from .name import LabelName
from .name import Name
from .name import OutlineImageName



"""
Functions associated with legacy python
"""


def cmp(a, b):
    return int((a > b)) - int((a < b))


def equals(a, b, encoding_a="utf-8", encoding_b="utf-8"):
    """
    Purpose: Used primarily for equality comparisons between bytes and str but can be used for any type

    :param a: any python type, including bytes
    :param b: any python type, including bytes
    :param encoding_a: str designating encoding format [default="utf-8"]
    :param encoding_b: str designating encoding format [default="utf-8"]
    :return: bool
    """
    if isinstance(a, bytes):
        a = a.decode(encoding_a)
    if isinstance(b, bytes):
        b = b.decode(encoding_b)
    return a == b


def convert_bytes_to_str(a, encoding="utf-8"):
    """
    Purpose: Converts the byte objects in the 1-D container into str (does not work for nested elements)
    :param a: iterable 1-D container
    :param encoding: str designating encoding format [default="utf-8"]
    :return: 1-D container with bytes converted to str
    """
    a_type = type(a)
    return a_type(map(lambda x: x.decode(encoding) if isinstance(x, bytes) else x, a))



import numpy as np

from cellprofiler_library.functions.segmentation import convert_dense_to_label_set
from cellprofiler_library.functions.segmentation import convert_sparse_to_ijv
from cellprofiler_library.functions.segmentation import indices_from_ijv
from cellprofiler_library.functions.segmentation import count_from_ijv
from cellprofiler_library.functions.segmentation import areas_from_ijv
from cellprofiler_library.functions.segmentation import convert_labels_to_dense
from cellprofiler_library.functions.segmentation import convert_ijv_to_sparse
from cellprofiler_library.functions.segmentation import make_rgb_outlines
from cellprofiler_library.functions.segmentation import find_label_overlaps
from cellprofiler_library.functions.segmentation import find_ijv_overlaps
from cellprofiler_library.functions.segmentation import center_of_labels_mass

from ._segmentation import Segmentation


class Objects:
    """Represents a segmentation of an image.

    IdentityPrimAutomatic produces three variants of its segmentation
    result. This object contains all three.

    There are three formats for segmentation, two of which support
    overlapping objects:

    get/set_segmented - Legacy, a single plane of labels that does not
                        support overlapping objects.

    (see cellprofiler_library.functions.segmentation._validate_labels)


    get_labels        - Supports overlapping objects, returns one or more planes
                        along with indices. A typical usage is to perform an
                        operation per-plane as if the objects did not overlap.

    (see cellprofiler_library.functions.segmentation.convert_dense_to_label_set)


    get/set_ijv       - Supports overlapping objects, returns a sparse
                        representation in which the first two columns are the
                        coordinates and the last is the object number. This
                        is efficient for doing things like calculating intensity
                        per-object.

    (see cellprofiler_library.functions.segmentation._validate_ijv)


    You can set one of the types and then get any of the types (except that
    get_segmented will raise an exception if objects overlap).
    """

    def __init__(self):
        self.__segmented = None
        self.__unedited_segmented = None
        self.__small_removed_segmented = None
        self.__parent_image = None

    @property
    def dimensions(self):
        if self.__parent_image:
            return self.__parent_image.dimensions

        shape = self.shape

        return len(shape)

    @property
    def volumetric(self):
        return self.dimensions == 3

    @property
    def masked(self):
        parent_image = self.parent_image

        assert parent_image is not None, "No parent image"

        return np.logical_and(self.segmented, parent_image.mask)

    @property
    def shape(self):
        dense, _ = self.__segmented.get_dense()

        if dense.shape[3] == 1:
            return dense.shape[-2:]

        return dense.shape[-3:]

    def get_segmented(self):
        """Get the de-facto segmentation of the image into objects: a matrix
        of object numbers.
        """
        return self.__segmentation_to_labels(self.__segmented)

    def set_segmented(self, labels):
        self.__segmented = self.__labels_to_segmentation(labels)

    segmented = property(get_segmented, set_segmented)

    @staticmethod
    def __labels_to_segmentation(labels):
        dense = convert_labels_to_dense(labels)
        return Segmentation(dense=dense)

    @staticmethod
    def __segmentation_to_labels(segmentation):
        assert isinstance(
            segmentation, Segmentation
        ), "Operation failed because objects were not initialized"

        dense, indices = segmentation.get_dense()

        assert (
            len(dense) == 1
        ), "Operation failed because objects overlapped. Please try with non-overlapping objects"

        if dense.shape[3] == 1:
            return dense.reshape(dense.shape[-2:])

        return dense.reshape(dense.shape[-3:])

    @property
    def indices(self):
        return indices_from_ijv(self.ijv, validate=False)

    @property
    def count(self):
        return count_from_ijv(self.ijv, validate=False)

    @property
    def areas(self):
        return areas_from_ijv(self.ijv, validate=False)

    def set_ijv(self, ijv, shape=None):
        """Set the segmentation to an IJV object format

        The ijv format is a list of i,j coordinates in slots 0 and 1
        and the label at the pixel in slot 2.
        """
        sparse = convert_ijv_to_sparse(ijv)

        if shape is not None:
            shape = (1, 1, 1, shape[0], shape[1])

        self.__segmented = Segmentation(sparse=sparse, shape=shape)

    def get_ijv(self):
        """Get the segmentation in IJV object format

        The ijv format is a list of i,j coordinates in slots 0 and 1
        and the label at the pixel in slot 2.
        """
        sparse = self.__segmented.sparse
        return convert_sparse_to_ijv(sparse, validate=False)

    ijv = property(get_ijv, set_ijv)

    def get_labels(self):
        """Get a set of labels matrices consisting of non-overlapping labels

        In IJV format, a single pixel might have multiple labels. If you
        want to use a labels matrix, you have an ambiguous situation and the
        resolution is to process separate labels matrices consisting of
        non-overlapping labels.

        returns a list of label matrixes and the indexes in each
        """
        dense, indices = self.__segmented.get_dense()

        return convert_dense_to_label_set(dense, indices=indices, validate=False)

    def has_unedited_segmented(self):
        """Return true if there is an unedited segmented matrix."""
        return self.__unedited_segmented is not None

    @property
    def unedited_segmented(self):
        """Get the segmentation of the image into objects, including junk that
        should be ignored: a matrix of object numbers.

        The default, if no unedited matrix is available, is the
        segmented labeling.
        """
        if self.__unedited_segmented is not None:
            return self.__segmentation_to_labels(self.__unedited_segmented)

        return self.segmented

    @unedited_segmented.setter
    def unedited_segmented(self, labels):
        self.__unedited_segmented = self.__labels_to_segmentation(labels)

    def has_small_removed_segmented(self):
        """Return true if there is a junk object matrix."""
        return self.__small_removed_segmented is not None

    @property
    def small_removed_segmented(self):
        """Get the matrix of segmented objects with the small objects removed

        This should be the same as the unedited_segmented label matrix with
        the small objects removed, but objects touching the sides of the image
        or the image mask still present.
        """
        if self.__small_removed_segmented is not None:
            return self.__segmentation_to_labels(self.__small_removed_segmented)

        return self.unedited_segmented

    @small_removed_segmented.setter
    def small_removed_segmented(self, labels):
        self.__small_removed_segmented = self.__labels_to_segmentation(labels)

    @property
    def parent_image(self):
        """The image that was analyzed to yield the objects.

        The image is an instance of CPImage which means it has the mask
        and crop mask.
        """
        return self.__parent_image

    @parent_image.setter
    def parent_image(self, parent_image):
        self.__parent_image = parent_image
        for segmentation in (
            self.__segmented,
            self.__small_removed_segmented,
            self.__unedited_segmented,
        ):
            if segmentation is not None and not segmentation.has_shape():
                shape = (
                    1,
                    1,
                    1,
                    parent_image.pixel_data.shape[0],
                    parent_image.pixel_data.shape[1],
                )
                segmentation.shape = shape

    @property
    def has_parent_image(self):
        """True if the objects were derived from a parent image

        """
        return self.__parent_image is not None

    def crop_image_similarly(self, image):
        """Crop an image in the same way as the parent image was cropped."""
        if image.shape == self.segmented.shape:
            return image
        if self.parent_image is None:
            raise ValueError("Images are of different size and no parent image")
        return self.parent_image.crop_image_similarly(image)

    def make_ijv_outlines(self, colors):
        """Make ijv-style color outlines

        Make outlines, coloring each object differently to distinguish between
        objects that might overlap.

        colors: a N x 3 color map to be used to color the outlines
        """
        return make_rgb_outlines(self.get_labels(), colors, validate=True)

    def relate_children(self, children):
        """Relate the object numbers in one label to the object numbers in another

        children - another "objects" instance: the labels of children within
                   the parent which is "self"

        Returns two 1-d arrays. The first gives the number of children within
        each parent. The second gives the mapping of each child to its parent's
        object number.
        """
        if self.volumetric:
            histogram = self.histogram_from_labels(self.segmented, children.segmented)
        else:
            histogram = self.histogram_from_ijv(self.ijv, children.ijv)

        return self.relate_histogram(histogram)

    def relate_labels(self, parent_labels, child_labels):
        """relate the object numbers in one label to those in another

        parent_labels - 2d label matrix of parent labels

        child_labels - 2d label matrix of child labels

        Returns two 1-d arrays. The first gives the number of children within
        each parent. The second gives the mapping of each child to its parent's
        object number.
        """
        histogram = self.histogram_from_labels(parent_labels, child_labels)
        return self.relate_histogram(histogram)

    @staticmethod
    def relate_histogram(histogram):
        """Return child counts and parents of children given a histogram

        histogram - histogram from histogram_from_ijv or histogram_from_labels
        """
        parent_count = histogram.shape[0] - 1

        parents_of_children = np.asarray(histogram.argmax(axis=0))
        if len(parents_of_children.shape) == 2:
            parents_of_children = np.squeeze(parents_of_children, axis=0)
        #
        # Create a histogram of # of children per parent
        children_per_parent = np.histogram(
            parents_of_children[1:], np.arange(parent_count + 2)
        )[0][1:]

        #
        # Make sure to remove the background elements at index 0
        #
        return children_per_parent, parents_of_children[1:]

    @staticmethod
    def histogram_from_labels(parent_labels, child_labels):
        """Find per pixel overlap of parent labels and child labels

        parent_labels - the parents which contain the children
        child_labels - the children to be mapped to a parent

        Returns a sparse matrix of overlap between each parent and child.
        Note that the first row and column are empty, as these
        correspond to parent and child labels of 0.
        """
        return find_label_overlaps(parent_labels, child_labels, validate=True)

    @staticmethod
    def histogram_from_ijv(parent_ijv, child_ijv):
        """Find per pixel overlap of parent labels and child labels,
        stored in ijv format.

        parent_ijv - the parents which contain the children
        child_ijv - the children to be mapped to a parent

        Returns a sparse matrix of overlap between each parent and child.
        Note that the first row and column are empty, as these
        correspond to parent and child labels of 0.
        """
        return find_ijv_overlaps(parent_ijv, child_ijv, validate=True)

    def fn_of_label_and_index(self, func):
        """Call a function taking a label matrix with the segmented labels

        function - should have signature like
                   labels - label matrix
                   index  - sequence of label indices documenting which
                            label indices are of interest
        """
        return func(self.segmented, self.indices)

    def fn_of_ones_label_and_index(self, func):
        """Call a function taking an image, a label matrix and an index with an image of all ones

        function - should have signature like
                   image  - image with same dimensions as labels
                   labels - label matrix
                   index  - sequence of label indices documenting which
                            label indices are of interest
        Pass this function an "image" of all ones, for instance to compute
        a center or an area
        """
        return func(np.ones(self.segmented.shape), self.segmented, self.indices)

    def center_of_mass(self):
        return center_of_labels_mass(self.segmented, validate=False)

    def overlapping(self):
        if not isinstance(self.__segmented, Segmentation):
            return False
        dense, indices = self.__segmented.get_dense()
        return len(dense) != 1



from cellprofiler_core.utilities.zmq.communicable.reply._reply import Reply


class LockStatusReply(Reply):
    """A reply to the LockStatusRequest

    self.uid - the unique ID of the locked file
    self.locked - true if locked, false if not
    """

    def __init__(self, uid, locked, **kwargs):
        Reply.__init__(self, **kwargs)
        self.uid = uid
        self.locked = locked



import os

from ._does_not_predicate import DoesNotPredicate
from ._does_predicate import DoesPredicate
from ._filter_predicate import FilterPredicate
from .._file_collection_display import FileCollectionDisplay
from ...utilities.image import is_image_extension

IS_TIF_PREDICATE = FilterPredicate(
    "istif",
    '"tif", "tiff", "ome.tif" or "ome.tiff"',
    lambda x: x.lower() in (".tif", ".tiff", ".ome.tif", ".ome.tiff"),
    [],
    doc="The extension is associated with TIFF image files",
)
IS_JPEG_PREDICATE = FilterPredicate(
    "isjpeg",
    '"jpg" or "jpeg"',
    lambda x: x.lower() in (".jpg", ".jpeg"),
    [],
    doc="The extension is associated with JPEG image files",
)
IS_PNG_PREDICATE = FilterPredicate(
    "ispng",
    '"png"',
    lambda x: x.lower() == ".png",
    [],
    doc="The extension is associated with PNG image files",
)
IS_IMAGE_PREDICATE = FilterPredicate(
    "isimage",
    "the extension of an image file",
    is_image_extension,
    [],
    "Is an extension commonly associated with image files",
)
IS_FLEX_PREDICATE = FilterPredicate(
    "isflex",
    '"flex"',
    lambda x: x.lower() == ".flex",
    [],
    doc="The extension is associated with .flex files",
)
IS_MOVIE_PREDICATE = FilterPredicate(
    "ismovie",
    '"mov" or "avi"',
    lambda x: x.lower() in (".mov", ".avi"),
    [],
    doc="The extension is associated with movie files",
)


class ExtensionPredicate(FilterPredicate):
    """A predicate that operates on file extensions"""

    def __init__(self):
        subpredicates = (
            IS_TIF_PREDICATE,
            IS_JPEG_PREDICATE,
            IS_PNG_PREDICATE,
            IS_IMAGE_PREDICATE,
            IS_FLEX_PREDICATE,
            IS_MOVIE_PREDICATE,
        )
        predicates = [
            DoesPredicate(subpredicates, "Is"),
            DoesNotPredicate(subpredicates, "Is not"),
        ]
        FilterPredicate.__init__(
            self,
            "extension",
            "Extension",
            self.fn_filter,
            predicates,
            doc="The rule applies to the file extension",
        )

    @staticmethod
    def fn_filter(node_type__modpath__module, *args):
        """The ExtensionPredicate filter function

        If the element is a file, try the different predicates on
        all possible extension parsings.
        """
        (node_type, modpath, module) = node_type__modpath__module
        if node_type == FileCollectionDisplay.NODE_DIRECTORY:
            return None
        elif isinstance(modpath[-1], tuple) and len(modpath[-1]) == 3:
            filename = modpath[-2]
        else:
            filename = modpath[-1]
        exts = ""
        while True:
            filename, ext = os.path.splitext(filename)
            if len(filename) == 0 or len(ext) == 0:
                return False
            exts = ext + exts
            if args[0](exts, *args[1:]):
                return True

    def test_valid(self, pipeline, *args):
        self((FileCollectionDisplay.NODE_FILE, ["/imaging", "test.tif"], None,), *args)



from ...utilities.zmq.communicable.request import AnalysisRequest


class ImageSetSuccess(AnalysisRequest):
    def __init__(self, analysis_id, image_set_number=None):
        AnalysisRequest.__init__(self, analysis_id, image_set_number=image_set_number)



from ._analysis_cancel import AnalysisCancel
from ._debug_complete import DebugComplete
from ._debug_waiting import DebugWaiting
from ._debug_waiting import DebugWaiting
from ._display import Display
from ._display_post_group import DisplayPostGroup
from ._display_post_run import DisplayPostRun
from ._exception_report import ExceptionReport
from ._initial_measurements import InitialMeasurements
from ._interaction import Interaction
from ._measurements_report import MeasurementsReport
from ._omero_login import OmeroLogin
from ._pipeline_preferences import PipelinePreferences
from ._shared_dictionary import SharedDictionary
from ._work import Work



import re


class FilterPredicate(object):
    def __init__(self, symbol, display_name, function, subpredicates, doc=None):
        self.symbol = symbol
        self.display_name = display_name
        self.function = function
        self.subpredicates = subpredicates
        self.doc = doc

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def test_valid(self, pipeline, *args):
        """Try running the filter on a test string"""
        self("", *args)

    @classmethod
    def encode_symbol(cls, symbol):
        """Escape encode an abritrary symbol name

        The parser needs to have special characters escaped. These are
        backslash, open and close parentheses, space and double quote.
        """
        return re.escape(symbol)

    @classmethod
    def decode_symbol(cls, symbol):
        """Decode an escape-encoded symbol"""
        s = ""
        in_escape = False
        for c in symbol:
            if in_escape:
                in_escape = False
                s += c
            elif c == "\\":
                in_escape = True
            else:
                s += c
        return s



from ._file_image import FileImage
from ._flex_frame_image import FlexFrameImage
from ._movie_frame_image import MovieFrameImage
from ._stk_frame_image import STKFrameImage
from .url import ColorImage
from .url import MaskImage
from .url import MonochromeImage
from .url import ObjectsImage
from .url import URLImage



import logging

import numpy

from ._grayscale_image import GrayscaleImage
from ._rgb_image import RGBImage
from .abstract_image import VanillaImage


LOGGER = logging.getLogger(__name__)

class ImageSet:
    """Represents the images for a particular iteration of a pipeline

    An image set is composed of one image provider per image in the set.
    The image provider loads or creates an image, given a dictionary of keys
    (which might represent things like the plate/well for the image set or the
    frame number in a movie, etc.)
    """

    def __init__(self, number, keys, legacy_fields):
        """Constructor:
        number = image set index
        keys = dictionary of key/value pairs that uniquely identify the image set
        """
        self.__image_providers = []
        self.__images = {}
        self.keys = keys
        self.number = number
        self.legacy_fields = legacy_fields
        self.image_number = number + 1

    def get_image(
        self,
        name,
        must_be_binary=False,
        must_be_color=False,
        must_be_grayscale=False,
        must_be_rgb=False,
    ):
        """Return the image associated with the given name

        name - name of the image within the image_set
        must_be_color - raise an exception if not a color image
        must_be_grayscale - raise an exception if not a grayscale image
        must_be_rgb - raise an exception if 2-d or if # channels not 3 or 4,
                      discard alpha channel.
        """
        name = str(name)
        if name not in self.__images:
            image = self.get_image_provider(name).provide_image(self)

        else:
            image = self.__images[name]

        if image.multichannel:
            if must_be_binary:
                raise ValueError("Image must be binary, but it was color")

            if must_be_grayscale:
                pd = image.pixel_data

                pd = pd.transpose(-1, *list(range(pd.ndim - 1)))

                if (
                    pd.shape[-1] >= 3
                    and numpy.all(pd[0] == pd[1])
                    and numpy.all(pd[0] == pd[2])
                ):
                    return GrayscaleImage(image)

                raise ValueError("Image must be grayscale, but it was color")

            if must_be_rgb:
                if image.pixel_data.shape[-1] not in (3, 4):
                    raise ValueError(
                        "Image must be RGB, but it had %d channels"
                        % image.pixel_data.shape[-1]
                    )

                if image.pixel_data.shape[-1] == 4:
                    LOGGER.warning("Discarding alpha channel.")

                    return RGBImage(image)

            return image

        if must_be_binary and image.pixel_data.dtype != bool:
            raise ValueError("Image was not binary")

        if must_be_grayscale and image.pixel_data.dtype.kind == "b":
            return GrayscaleImage(image)

        if must_be_rgb:
            raise ValueError("Image must be RGB, but it was grayscale")

        if must_be_color:
            raise ValueError("Image must be color, but it was grayscale")

        return image

    @property
    def providers(self):
        """The list of providers (populated during the image discovery phase)"""
        return tuple(self.__image_providers)
    
    def add_provider(self, provider):
        self.__image_providers.append(provider)

    def get_image_provider(self, name):
        """Get a named image provider

        name - return the image provider with this name
        """
        providers = [x for x in self.__image_providers if x.name == name]
        assert len(providers) > 0, "No provider of the %s image" % name
        assert len(providers) == 1, "More than one provider of the %s image" % name
        return providers[0]

    def remove_image_provider(self, name):
        """Remove a named image provider

        name - the name of the provider to remove
        """
        self.__image_providers = [x for x in self.__image_providers if x.name != name]

    def clear_image(self, name):
        """Remove the image memory associated with a provider

        name - the name of the provider
        """
        self.get_image_provider(name).release_memory()
        if name in self.__images:
            del self.__images[name]

    @property
    def names(self):
        """Get the image provider names
        """
        return [provider.name for provider in self.providers]

    def add(self, name, image):
        old_providers = [
            provider for provider in self.providers if provider.name == name
        ]
        if len(old_providers) > 0:
            self.clear_image(name)
        for provider in old_providers:
            self.remove_image_provider(provider.name)
        provider = VanillaImage(name, image)
        self.add_provider(provider)



from ...utilities.zmq.communicable.request import AnalysisRequest


class ExceptionReport(AnalysisRequest):
    def __init__(
        self,
        analysis_id,
        image_set_number,
        module_name,
        exc_type,
        exc_message,
        exc_traceback,
        filename,
        line_number,
    ):
        AnalysisRequest.__init__(
            self,
            analysis_id,
            image_set_number=image_set_number,
            module_name=module_name,
            exc_type=exc_type,
            exc_message=exc_message,
            exc_traceback=exc_traceback,
            filename=filename,
            line_number=line_number,
        )

    def __str__(self):
        return "(Worker) %s: %s" % (self.exc_type, self.exc_message)



#TODO: unimplemented until CellProfiler/CellProfiler#4684 is resolved

K_OMERO_SERVER = None
K_OMERO_PORT = None
K_OMERO_USER = None
K_OMERO_SESSION_ID = None
K_OMERO_CONFIG_FILE = None

def clear_image_reader_cache():
    raise RuntimeError("unimplemented")

def set_omero_login_hook(omero_login):
    raise RuntimeError("unimplemented")

def get_omero_credentials():
    raise RuntimeError("unimplemented")

def use_omero_credentials(credentials):
    raise RuntimeError("unimplemented")



import errno
import json
import logging
import os
import queue
import sys
import threading
import uuid

import numpy
import zmq

import cellprofiler_core.utilities.grid
from cellprofiler_core.utilities.zmq._boundary import Boundary
from cellprofiler_core.utilities.zmq.communicable.reply import LockStatusReply, Reply
from cellprofiler_core.utilities.zmq.communicable.request import (
    LockStatusRequest,
    Request,
)
from ._event import PollTimeoutException


LOGGER = logging.getLogger(__name__)

NOTIFY_SOCKET_ADDR = "inproc://BoundaryNotifications"
SD_KEY_DICT = "__keydict__"


class CPJSONEncoder(json.JSONEncoder):
    def encode(self, obj):
        def hint_tuples(item):
            if isinstance(item, tuple):
                return {'__tuple__': True, 'items': item}
            if isinstance(item, list):
                return [hint_tuples(e) for e in item]
            if isinstance(item, dict):
                return {key: hint_tuples(value) for key, value in item.items()}
            else:
                return item

        return super().encode(hint_tuples(obj))

def make_CP_fallback_encoder(buffers):
    """create an encoder for CellProfiler data and numpy arrays (which will be
    stored in the input argument)"""

    def encoder(data, buffers=buffers):
        if isinstance(data, numpy.ndarray):
            #
            # Maybe it's nice to save memory by converting 64-bit to 32-bit
            # but the purpose here is to fix a bug on the Mac where a
            # 32-bit worker gets a 64-bit array or unsigned 32-bit array,
            # tries to use it for indexing and fails because the integer
            # is wider than a 32-bit pointer
            #
            info32 = numpy.iinfo(numpy.int32)
            if (
                data.dtype.kind == "i"
                and data.dtype.itemsize > 4
                or data.dtype.kind == "u"
                and data.dtype.itemsize >= 4
            ):
                if numpy.prod(data.shape) == 0 or (
                    numpy.min(data) >= info32.min and numpy.max(data) <= info32.max
                ):
                    data = data.astype(numpy.int32)
            idx = len(buffers)
            buffers.append(numpy.ascontiguousarray(data))
            dtype = str(data.dtype)
            return {
                "__ndarray__": True,
                "dtype": str(data.dtype),
                "shape": data.shape,
                "idx": idx,
            }
        if isinstance(data, numpy.generic):
            # http://docs.scipy.org/doc/numpy/reference/arrays.scalars.html
            return data.astype(object)
        if isinstance(data, cellprofiler_core.utilities.grid.Grid):
            d = data.serialize()
            d["__CPGridInfo__"] = True
            return d
        if isinstance(data, memoryview):
            # arbitrary data
            idx = len(buffers)
            buffers.append(data)
            return {"__buffer__": True, "idx": idx}
        raise TypeError("%r of type %r is not JSON serializable" % (data, type(data)))

    return encoder


def make_CP_decoder(buffers):
    def decoder(dct, buffers=buffers):
        if '__tuple__' in dct:
            return tuple(dct['items'])
        if "__ndarray__" in dct:
            buf = memoryview(buffers[dct["idx"]])
            shape = dct["shape"]
            dtype = dct["dtype"]
            if numpy.prod(shape) == 0:
                return numpy.zeros(shape, dtype)
            return numpy.frombuffer(buf, dtype=dtype).reshape(shape).copy()
        if "__buffer__" in dct:
            return memoryview(buffers[dct["idx"]])
        if "__CPGridInfo__" in dct:
            grid = cellprofiler_core.utilities.grid.Grid()
            grid.deserialize(dct)
            return grid
        return dct

    return decoder


def make_sendable_dictionary(d):
    """Make a dictionary that passes muster with JSON"""
    result = {}
    fake_key_idx = 1
    for k, v in list(d.items()):
        if (isinstance(k, str) and k.startswith("_")) or callable(d[k]):
            continue
        if isinstance(v, dict):
            v = make_sendable_dictionary(v)
        elif isinstance(v, (list, tuple)):
            v = make_sendable_sequence(v)
        if not isinstance(k, str):
            if SD_KEY_DICT not in result:
                result[SD_KEY_DICT] = {}
            fake_key = "__%d__" % fake_key_idx
            fake_key_idx += 1
            result[SD_KEY_DICT][fake_key] = k
            result[fake_key] = v
        else:
            result[k] = v
    return result


def make_sendable_sequence(l):
    """Make a list that passes muster with JSON"""
    result = []
    for v in l:
        if isinstance(v, (list, tuple)):
            result.append(make_sendable_sequence(v))
        elif isinstance(v, dict):
            result.append(make_sendable_dictionary(v))
        else:
            result.append(v)
    if isinstance(l, tuple):
        return tuple(result)
    else: # else return as list
        return result


def decode_sendable_dictionary(d):
    """Decode the dictionary encoded by make_sendable_dictionary"""
    result = {}
    for k, v in list(d.items()):
        if k == SD_KEY_DICT:
            continue
        if isinstance(v, dict):
            v = decode_sendable_dictionary(v)
        elif isinstance(v, list):
            v = decode_sendable_sequence(v, list)
        if k.startswith("__") and k.endswith("__"):
            k = d[SD_KEY_DICT][k]
            if isinstance(k, list):
                k = decode_sendable_sequence(k, tuple)
        result[k] = v
    return result


def decode_sendable_sequence(l, desired_type):
    """Decode a tuple encoded by make_sendable_sequence"""
    result = []
    for v in l:
        if isinstance(v, dict):
            result.append(decode_sendable_dictionary(v))
        elif isinstance(v, (list, tuple)):
            result.append(decode_sendable_sequence(v, desired_type))
        else:
            result.append(v)
    return result if isinstance(result, desired_type) else desired_type(result)


def json_encode(o):
    """Encode an object as a JSON string

    o - object to encode

    returns a 2-tuple of json-encoded object + buffers of binary stuff
    """
    sendable_dict = make_sendable_dictionary(o)

    # replace each buffer with its metadata, and send it separately
    buffers = []
    fallback_encoder = make_CP_fallback_encoder(buffers)
    json_str = json.dumps(sendable_dict, cls=CPJSONEncoder, default=fallback_encoder)
    return json_str, buffers


def json_decode(json_str, buffers):
    """Decode a JSON-encoded string

    json_str - the JSON string

    buffers - buffers of binary data to feed into the decoder of special cases

    return the decoded dictionary
    """
    decoder = make_CP_decoder(buffers)
    attribute_dict = json.loads(json_str, object_hook=decoder)
    return decode_sendable_dictionary(attribute_dict)


the_boundary = None


def start_boundary():
    global the_boundary
    if the_boundary is None:
        the_boundary = Boundary("tcp://127.0.0.1")
    return the_boundary


def get_announcer_address():
    return start_boundary().announce_address


def register_analysis(analysis_id, upward_queue):
    """Register for all analysis request messages with the given ID

    analysis_id - the analysis ID present in every AnalysisRequest

    upward_queue - requests are placed on this queue

    upward_cv - the condition variable used to signal the queue's thread

    returns the boundary singleton.
    """
    global the_boundary
    start_boundary()
    the_boundary.register_analysis(analysis_id, upward_queue)
    return the_boundary


def join_to_the_boundary():
    """Send a stop signal to the boundary thread and join to it"""
    global the_boundary
    if the_boundary is not None:
        the_boundary.join()
        the_boundary = None


__lock_queue: queue.Queue = queue.Queue()
__lock_thread = None
LOCK_REQUEST = "Lock request"
UNLOCK_REQUEST = "Unlock request"
UNLOCK_OK = "OK"


def start_lock_thread():
    """Start the thread that handles file locking"""
    global __lock_thread
    if __lock_thread is not None:
        return
    the_boundary.register_request_class(LockStatusRequest, __lock_queue)

    def lock_thread_fn():
        global __lock_thread
        locked_uids = {}
        locked_files = {}
        while True:
            msg = __lock_queue.get()
            boundary = msg[0]
            if msg[1] == Boundary.NOTIFY_STOP:
                msg[2].put(__lock_thread)
                break
            elif msg[1] == Boundary.NOTIFY_REQUEST:
                request = msg[2]
                assert isinstance(request, LockStatusRequest)
                assert isinstance(boundary, Boundary)
                LOGGER.info("Received lock status request for %s" % request.uid)
                reply = LockStatusReply(request.uid, request.uid in locked_uids)
                if reply.locked:
                    LOGGER.info(
                        "Denied lock request for %s" % locked_uids[request.uid]
                    )
                boundary.enqueue_reply(request, reply)
            elif msg[1] == LOCK_REQUEST:
                uid, path = msg[2]
                locked_uids[uid] = path
                locked_files[path] = uid
                msg[3].put("OK")
            elif msg[1] == UNLOCK_REQUEST:
                try:
                    uid = locked_files[msg[2]]
                    del locked_uids[uid]
                    del locked_files[msg[2]]
                    msg[3].put("OK")
                except Exception as e:
                    msg[3].put(e)
        __lock_thread = None
        LOGGER.info("Exiting the lock thread")

    __lock_thread = threading.Thread(target=lock_thread_fn, name="FileLock thread")
    __lock_thread.start()


def get_lock_path(path):
    """Return the path to the lockfile"""
    pathpart, filepart = os.path.split(path)
    return os.path.join(pathpart, "." + filepart + ".lock")


def lock_file(path, timeout=3):
    """Lock a file

    path - path to the file

    timeout - timeout in seconds when waiting for announcement

    returns True if we obtained the lock, False if the file is already owned.
    """
    lock_path = get_lock_path(path)
    start_boundary()
    uid = uuid.uuid4().hex
    try:
        fd = os.open(lock_path, os.O_CREAT | os.O_EXCL | os.O_RDWR)
        with os.fdopen(fd, "a") as f:
            f.write(the_boundary.external_request_address + "\n" + uid)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
        LOGGER.info("Lockfile for %s already exists - contacting owner" % path)
        with open(lock_path, "r") as f:
            remote_address = f.readline().strip()
            remote_uid = f.readline().strip()
        if len(remote_address) > 0 and len(remote_uid) > 0:
            LOGGER.info("Owner is %s" % remote_address)
            request_socket = the_boundary.zmq_context.socket(zmq.REQ)
            request_socket.setsockopt(zmq.LINGER, 0)
            assert isinstance(request_socket, zmq.Socket)
            request_socket.connect(remote_address)

            lock_request = LockStatusRequest(remote_uid)
            lock_request.send_only(request_socket)
            poller = zmq.Poller()
            poller.register(request_socket, zmq.POLLIN)
            keep_polling = True
            while keep_polling:
                keep_polling = False
                for socket, status in poller.poll(timeout * 1000):
                    keep_polling = True
                    if socket == request_socket and status == zmq.POLLIN:
                        lock_response = lock_request.recv(socket)
                        if isinstance(lock_response, LockStatusReply):
                            if lock_response.locked:
                                LOGGER.info("%s is locked" % path)
                                return False
                            keep_polling = False
        #
        # Fall through if we believe that the other end is dead
        #
        with open(lock_path, "w") as f:
            f.write(the_boundary.request_address + "\n" + uid)
    #
    # The coast is clear to lock
    #
    q = queue.Queue()
    start_lock_thread()
    __lock_queue.put((None, LOCK_REQUEST, (uid, path), q))
    q.get()
    return True


def unlock_file(path):
    """Unlock the file at the given path"""
    if the_boundary is None:
        return
    q = queue.Queue()
    start_lock_thread()
    __lock_queue.put((None, UNLOCK_REQUEST, path, q))
    result = q.get()
    if result != UNLOCK_OK:
        raise result
    lock_path = get_lock_path(path)
    os.remove(lock_path)


context = zmq.Context()


def subproc():
    address = sys.argv[sys.argv.index("subproc") + 1]
    mysock = context.socket(zmq.REQ)
    mysock.connect(address)
    req = Request(
        this="is", a="test", b=5, c=1.3, d=numpy.arange(10), e=[{"q": numpy.arange(5)}],
    )
    rep = req.send(mysock)
    print("subproc received", rep, rep.__dict__)
    rep = rep.reply(Reply(msg="FOO"), please_reply=True)
    print("subproc received", rep, rep.__dict__)



class ChangeSettingEvent:
    """Abstract class representing either the event that a setting will be
    changed or has been changed

    """

    def __init__(self, old_value, new_value):
        self.__old_value = old_value
        self.__new_value = new_value

    def get_old_value(self):
        return self.__old_value

    old_value = property(get_old_value)

    def get_new_value(self):
        return self.__new_value

    new_value = property(get_new_value)



import json

from ._setting import Setting


class DataTypes(Setting):
    """The DataTypes setting assigns data types to measurement names

    Imported or extracted metadata might be textual or numeric and
    that interpretation should be up to the user. This setting lets
    the user pick the data type for their metadata.
    """

    DT_TEXT = "text"
    DT_INTEGER = "integer"
    DT_FLOAT = "float"
    DT_NONE = "none"

    def __init__(self, text, value="{}", name_fn=None, *args, **kwargs):
        """Initializer

        text - description of the setting

        value - initial value (a json-encodable key/value dictionary)

        name_fn - a function that returns the current list of feature names
        """
        super(DataTypes, self).__init__(text, value, *args, **kwargs)

        self.__name_fn = name_fn

    def get_data_types(self):
        """Get a dictionary of the data type for every name

        Using the name function, if present, create a dictionary of name
        to data type (DT_TEXT / INTEGER / FLOAT / NONE)
        """
        result = json.loads(self.value_text)
        if self.__name_fn is not None:
            for name in self.__name_fn():
                if name not in result:
                    result[name] = self.DT_TEXT
        return result

    @staticmethod
    def decode_data_types(s):
        return json.loads(s)

    @staticmethod
    def encode_data_types(d):
        """Encode a data type dictionary as a potential value for this setting"""
        return json.dumps(d)



import os.path
import urllib.parse

from ..module import Module
from ..pipeline import ImagePlane as ImagePlane
from ..utilities.pathname import pathname2url


class SettingValidation(Module):
    """A fake module for setting validation"""

    @staticmethod
    def get_image_plane_details(modpath):
        if modpath[0] in ("http", "https", "ftp", "s3", "gs"):
            if len(modpath) == 1:
                return modpath[0] + ":"
            elif len(modpath) == 2:
                return modpath[0] + ":" + modpath[1]
            else:
                return (
                    modpath[0]
                    + ":"
                    + modpath[1]
                    + "/"
                    + "/".join([urllib.parse.quote(part) for part in modpath[2:]])
                )

        path = os.path.join(*modpath)

        url = pathname2url(path)

        return ImagePlane(ImageFile(url))



UIC1_TAG = 33628
UIC2_TAG = 33629
UIC3_TAG = 33630
UIC4_TAG = 33631
C_MD5_DIGEST = "MD5Digest"
C_SCALING = "Scaling"
C_HEIGHT = "Height"
C_WIDTH = "Width"
MS_EXACT_MATCH = "Text-Exact match"
MS_REGEXP = "Text-Regular expressions"
MS_ORDER = "Order"
FF_INDIVIDUAL_IMAGES = "individual images"
FF_STK_MOVIES = "stk movies"
FF_AVI_MOVIES = "avi,mov movies"
FF_AVI_MOVIES_OLD = ["avi movies"]
FF_OTHER_MOVIES = "tif,tiff,flex,zvi movies"
FF_OTHER_MOVIES_OLD = ["tif,tiff,flex movies", "tif,tiff,flex movies, zvi movies"]
IO_IMAGES = "Images"
IO_OBJECTS = "Objects"
IO_ALL = (IO_IMAGES, IO_OBJECTS)
IMAGE_FOR_OBJECTS_F = "IMAGE_FOR_%s"
SUPPORTED_IMAGE_EXTENSIONS = {
    ".arg",
    ".bmp",
    ".bufr",
    ".bw",
    ".cur",
    ".dcx",
    ".dib",
    ".emf",
    ".eps",
    ".fit",
    ".fits",
    ".flc",
    ".fli",
    ".fpx",
    ".gbr",
    ".gif",
    ".grib",
    ".h5",
    ".hdf",
    ".icns",
    ".ico",
    ".iim",
    ".im",
    ".jfif",
    ".jpe",
    ".jpeg",
    ".jpg",
    ".mic",
    ".mpeg",
    ".mpg",
    ".msp",
    ".palm",
    ".pbm",
    ".pcd",
    ".pcx",
    ".pdf",
    ".pgm",
    ".png",
    ".ppm",
    ".ps",
    ".psd",
    ".ras",
    ".rgb",
    ".rgba",
    ".sgi",
    ".tga",
    ".tif",
    ".tiff",
    ".wmf",
    ".xbm",
    ".xpm",
    ".zarr",
}
SUPPORTED_MOVIE_EXTENSIONS = {
    ".avi",
    ".flex",
    ".mov",
    ".mpeg",
    ".stk",
    ".tif",
    ".tiff",
    ".zvi",
}

# See https://docs.openmicroscopy.org/bio-formats/latest/supported-formats.html
ALL_BIOFORMATS_EXTENSIONS = (
    ".1sc", ".2fl", ".acff", ".afi", ".afm", ".aim", ".al3d", ".ali",
    ".am", ".amiramesh", ".apl", ".arf", ".avi", ".bif", ".bin", ".bip",
    ".bmp", ".btf", ".c01", ".cfg", ".ch5", ".cif", ".cr2", ".crw",
    ".cxd", ".czi", ".dat", ".dcm", ".dib", ".dicom", ".dm2", ".dm3",
    ".dm4", ".dti", ".dv", ".eps", ".epsi", ".exp", ".fdf", ".fff",
    ".ffr", ".fits", ".flex", ".fli", ".frm", ".gel", ".gif", ".grey",
    ".h5", ".hdf", ".hdr", ".hed", ".his", ".htd", ".html", ".hx", ".i2i",
    ".ics", ".ids", ".im3", ".img", ".ims", ".inr", ".ipl", ".ipm", ".ipw",
    ".j2k", ".jp2", ".jpeg", ".jpf", ".jpg", ".jpk", ".jpx", ".klb",
    ".l2d", ".labels", ".lei", ".lif", ".liff", ".lim", ".lms", ".lsm",
    ".map", ".mdb", ".mea", ".mnc", ".mng", ".mod", ".mov", ".mrc", ".mrcs",
    ".mrw", ".msr", ".mtb", ".mvd2", ".naf", ".nd", ".nd2", ".ndpi", ".ndpis",
    ".nef", ".nhdr", ".nii", ".nii.gz", ".nrrd", ".obf", ".obsep", ".oib",
    ".oif", ".oir", ".ome", ".ome.btf", ".ome.tf2", ".ome.tf8", ".ome.tif",
    ".ome.tiff", ".ome.xml", ".par", ".pbm", ".pcoraw", ".pcx", ".pds",
    ".pgm", ".pic", ".pict", ".png", ".pnl", ".ppm", ".pr3", ".ps", ".psd",
    ".qptiff", ".r3d", ".raw", ".rcpnl", ".rec", ".res", ".scn", ".sdt",
    ".seq", ".sif", ".sld", ".sm2", ".sm3", ".spc", ".spe", ".spi", ".st",
    ".stk", ".stp", ".svs", ".sxm", ".tc.", ".tf2", ".tf8", ".tfr", ".tga",
    ".tif", ".tiff", ".tnb", ".top", ".txt", ".v", ".vff", ".vms", ".vsi",
    ".vws", ".wat", ".wlz", ".wpi", ".xdce", ".xml", ".xqd", ".xqf", ".xv",
    ".xys", ".zfp", ".zfr", ".zvi", ".ets"
)

DISALLOWED_BIOFORMATS_EXTENSIONS = {
    ".cfg",
    ".csv",
    ".eps",
    ".epsi",
    ".htm",
    ".html",
    ".inf",
    ".log",
    ".ps",
    ".txt",
    ".wav",
    ".xml",
    ".zip"
}

BIOFORMATS_IMAGE_EXTENSIONS = set(ALL_BIOFORMATS_EXTENSIONS) - DISALLOWED_BIOFORMATS_EXTENSIONS

ALL_IMAGE_EXTENSIONS = SUPPORTED_IMAGE_EXTENSIONS.union(BIOFORMATS_IMAGE_EXTENSIONS)


FF = [FF_INDIVIDUAL_IMAGES, FF_STK_MOVIES, FF_AVI_MOVIES, FF_OTHER_MOVIES]
M_NONE = "None"
M_FILE_NAME = "File name"
M_PATH = "Path"
M_BOTH = "Both"
M_Z = "Z"
M_T = "T"
C_SERIES = "Series"
C_FRAME = "Frame"
P_IMAGES = "LoadImagesImageProvider"
V_IMAGES = 1
P_MOVIES = "LoadImagesMovieProvider"
V_MOVIES = 2
P_FLEX = "LoadImagesFlexFrameProvider"
V_FLEX = 1
I_INTERLEAVED = "Interleaved"
I_SEPARATED = "Separated"
SUB_NONE = "None"
SUB_ALL = "All"
SUB_SOME = "Some"
FILE_SCHEME = "file:"
PASSTHROUGH_SCHEMES = ["http", "https", "ftp", "s3", "gs"]

CT_GRAYSCALE = "Grayscale"
CT_COLOR = "Color"
CT_MASK = "Mask"
CT_OBJECTS = "Objects"
CT_FUNCTION = "Function"

MD_PLANAR = "Planar"
MD_SIZE_S = "SizeS"
MD_SIZE_C = "SizeC"
MD_SIZE_Z = "SizeZ"
MD_SIZE_T = "SizeT"
MD_SIZE_X = "SizeX"
MD_SIZE_Y = "SizeY"
MD_SERIES_NAME = "SeriesNames"
MD_SIZE_KEYS = {MD_SIZE_S, MD_SIZE_X, MD_SIZE_Y, MD_SIZE_Z, MD_SIZE_C, MD_SIZE_T}

NO_RESCALE = (None, None)



from .communicable import Communicable
from .communicable.reply.upstream_exit import BoundaryExited


class AnalysisContext:
    """The analysis context holds the pieces needed to route analysis requests"""

    def __init__(self, analysis_id, upq, lock):
        self.lock = lock
        self.analysis_id = analysis_id
        self.upq = upq
        self.cancelled = False
        # A map of requests pending to the closure that can be used to
        # reply to the request
        self.reqs_pending = set()

    def reply(self, req, rep):
        """Reply to a AnalysisRequest with this analysis ID

        rep - the intended reply

        Returns True if the intended reply was sent, returns False
        if BoundaryExited was sent instead.

        Always executed on the boundary thread.
        """
        with self.lock:
            if self.cancelled:
                return False
            if req in self.reqs_pending:
                Communicable.reply(req, rep)
                self.reqs_pending.remove(req)
            return True

    def enqueue(self, req):
        """Enqueue a request on the upward queue

        req - request to be enqueued. The enqueue should be done before
              req.reply is replaced.

        returns True if the request was enqueued, False if the analysis
        has been cancelled. It is up to the caller to send a BoundaryExited
        reply to the request.

        Always executes on the boundary thread.
        """
        with self.lock:
            if not self.cancelled:
                assert req not in self.reqs_pending
                self.reqs_pending.add(req)
                self.upq.put(req)
                return True
            else:
                Communicable.reply(req, BoundaryExited())
                return False

    def cancel(self):
        """Cancel this analysis

        All analysis requests will receive BoundaryExited() after this
        method returns.
        """
        with self.lock:
            if self.cancelled:
                return
            self.cancelled = True
            self.upq = None

    def handle_cancel(self):
        """Handle a cancel in the boundary thread.

        Take care of workers expecting replies.
        """
        with self.lock:
            for req in list(self.reqs_pending):
                Communicable.reply(req, BoundaryExited())
            self.reqs_pending = set()



from ._subscriber_multichoice import SubscriberMultiChoice


class ImageNameSubscriberMultiChoice(SubscriberMultiChoice):
    """A multi-choice setting that displays images

    This setting displays a list of images taken from ImageNameProviders.
    """

    def __init__(self, text, value=None, *args, **kwargs):
        super(ImageNameSubscriberMultiChoice, self).__init__(
            text, "imagegroup", value, *args, **kwargs
        )



from ._filter_predicate import FilterPredicate


class CompoundFilterPredicate(FilterPredicate):
    def test_valid(self, pipeline, *args):
        for subexp in args:
            subexp[0].test_valid(pipeline, *subexp[1:])



import os
import scyjava
import numpy as np
import logging

from . import omexml
from ..utilities.java import jimport


LOGGER = logging.getLogger(__name__)

p2j = lambda v: scyjava.to_java(v)

def write_image(
    pathname,
    pixels,
    pixel_type,
    c,
    z,
    t,
    size_c,
    size_z,
    size_t,
    channel_names
):
    # https://www.javadoc.io/doc/org.openmicroscopy/ome-common/5.3.2/loci/common/services/ServiceFactory.html
    OMEXMLServiceFactory = jimport("loci.common.services.ServiceFactory")
    # https://javadoc.scijava.org/Bio-Formats/loci/formats/services/OMEXMLService.html
    OMEXMLService = jimport("loci.formats.services.OMEXMLService")
    # https://javadoc.scijava.org/Bio-Formats/loci/formats/ImageWriter.html
    ImageWriter = jimport("loci.formats.ImageWriter")
    # https://www.javadoc.io/static/org.openmicroscopy/ome-xml/6.3.1/ome/xml/meta/IMetadata.html
    IMetadata = jimport("loci.formats.meta.IMetadata")

    DimensionsOrder = jimport("ome.xml.model.enums.DimensionOrder")
    PixelType = jimport("ome.xml.model.enums.PixelType")
    PositiveInteger = jimport("ome.xml.model.primitives.PositiveInteger")

    omexml_service = OMEXMLServiceFactory().getInstance(OMEXMLService)
    # https://www.javadoc.io/static/org.openmicroscopy/ome-xml/6.3.1/ome/xml/meta/OMEXMLMetadata.html
    # https://www.javadoc.io/static/org.openmicroscopy/ome-xml/6.3.1/ome/xml/meta/MetadataStore.html
    metadata = omexml_service.createOMEXMLMetadata()
    metadata.createRoot()

    metadata.setImageName(os.path.split(pathname)[1], 0)
    metadata.setPixelsSizeX(PositiveInteger(p2j(pixels.shape[1])), 0)
    metadata.setPixelsSizeY(PositiveInteger(p2j(pixels.shape[0])), 0)
    metadata.setPixelsSizeC(PositiveInteger(p2j(size_c)), 0)
    metadata.setPixelsSizeZ(PositiveInteger(p2j(size_z)), 0)
    metadata.setPixelsSizeT(PositiveInteger(p2j(size_t)), 0)
    metadata.setPixelsBinDataBigEndian(True, 0, 0)
    metadata.setPixelsDimensionOrder(DimensionsOrder.XYCTZ, 0)
    metadata.setPixelsType(PixelType.fromString(pixel_type), 0)


    if pixels.ndim == 3:
        metadata.setPixelsSizeC(PositiveInteger(p2j(pixels.shape[2])), 0)
        metadata.setChannelSamplesPerPixel(PositiveInteger(p2j(pixels.shape[2])), 0, 0)
        omexml_service.populateOriginalMetadata(metadata, "SamplesPerPixel", str(pixels.shape[2]))
        # omexml.structured_annotations.add_original_metadata(
        #     ome.OM_SAMPLES_PER_PIXEL, str(pixels.shape[2]))
    elif size_c > 1:
        # meta.channel_count = size_c <- cant find
        metadata.setPixelsSizeC(PositiveInteger(p2j(pixels.shape[2])), 0)
        omexml_service.populateOriginalMetadata(metadata, "SamplesPerPixel", str(pixels.shape[2]))
    
    metadata.setImageID("Image:0", 0)
    metadata.setPixelsID("Pixels:0", 0)

    for i in range(size_c):
        metadata.setChannelID(f"Channel:0:{i}", 0, i)
        metadata.setChannelSamplesPerPixel(PositiveInteger(p2j(1)), 0, i)
    
    index = c + size_c * z + size_c * size_z * t
    pixel_buffer = convert_pixels_to_buffer(pixels, pixel_type)


    writer = ImageWriter()
    writer.setMetadataRetrieve(metadata)
    writer.setId(pathname)
    writer.setInterleaved(True)
    writer.saveBytes(index, pixel_buffer)
    writer.close()

def convert_pixels_to_buffer(pixels, pixel_type):
    '''Convert the pixels in the image into a buffer of the right pixel type

    pixels - a 2d monochrome or color image

    pixel_type - one of the OME pixel types

    returns a 1-d byte array
    '''
    if pixel_type == omexml.PT_UINT8:
        as_dtype = np.uint8
    elif pixel_type == omexml.PT_UINT16:
        as_dtype = "<u2"
    elif pixel_type == omexml.PT_FLOAT:
        as_dtype = "<f4"
    else:
        raise NotImplementedError("Unsupported pixel type: %d" % pixel_type)

    return np.frombuffer(np.ascontiguousarray(pixels, as_dtype).data, np.uint8)




from ._does_not_predicate import DoesNotPredicate
from ._does_predicate import DoesPredicate
from ._filter import LITERAL_PREDICATE
from ._filter_predicate import FilterPredicate
from .._file_collection_display import FileCollectionDisplay
from ...pipeline import ImageFile
from ...pipeline import ImagePlane as ImagePlane


class MetadataPredicate(FilterPredicate):
    """A predicate that compares an ifd against a metadata key and value"""

    SYMBOL = "metadata"

    def __init__(self, display_name, display_fmt="%s", **kwargs):
        subpredicates = [
            DoesPredicate([]),
            DoesNotPredicate([]),
        ]

        super(self.__class__, self).__init__(
            self.SYMBOL,
            display_name,
            MetadataPredicate.do_filter,
            subpredicates,
            **kwargs,
        )
        self.display_fmt = display_fmt

    def set_metadata_keys(self, keys):
        """Define the possible metadata keys to be matched against literal values

        keys - a list of keys
        """
        sub_subpredicates = [
            FilterPredicate(
                key,
                self.display_fmt % key,
                lambda plane, match, key=key: str(plane.get_metadata(key)) == match,
                [LITERAL_PREDICATE],
            )
            for key in keys
        ]
        #
        # The subpredicates are "Does" and "Does not", so we add one level
        # below that.
        #
        for subpredicate in self.subpredicates:
            subpredicate.subpredicates = sub_subpredicates

    @classmethod
    def do_filter(cls, arg, *vargs):
        """Perform the metadata predicate's filter function

        The metadata predicate has subpredicates that look up their
        metadata key in the ipd and compare it against a literal.
        """
        node_type, modpath, plane = arg
        return vargs[0](plane, *vargs[1:])

    def test_valid(self, pipeline, *args):
        print("Testing validity of a metadata predicate using a fake image plane")

        class FakeModpathResolver(object):
            """Resolve one modpath to one ipd"""

            def __init__(self, modpath, ipd):
                self.modpath = modpath
                self.ipd = ipd

            def get_image_plane_details(self, modpath):
                assert len(modpath) == len(self.modpath)
                assert all([m1 == m2 for m1, m2 in zip(self.modpath, modpath)])
                return self.ipd

        modpath = ["imaging", "image.png"]
        plane = ImagePlane(ImageFile("/imaging/image.png"), None, None, None)
        self(
            (
                FileCollectionDisplay.NODE_IMAGE_PLANE,
                modpath,
                FakeModpathResolver(modpath, plane),
            ),
            *args,
        )



import skimage.morphology

from ._setting import Setting
from ._validation_error import ValidationError


class StructuringElement(Setting):
    def __init__(
        self,
        text="Structuring element",
        value="Disk,1",
        allow_planewise=False,
        *args,
        **kwargs,
    ):
        self.__allow_planewise = allow_planewise

        super(StructuringElement, self).__init__(text, value, *args, **kwargs)

    @staticmethod
    def get_choices():
        return ["Ball", "Cube", "Diamond", "Disk", "Octahedron", "Square", "Star"]

    def get_value(self):
        return getattr(skimage.morphology, self.shape)(self.size)

    def set_value(self, value):
        self.value_text = value

    @property
    def shape(self):
        return str(self.value_text.split(",")[0]).lower()

    @shape.setter
    def shape(self, value):
        self.value_text = ",".join((value, str(self.size)))

    @property
    def size(self):
        _, size = self.value_text.split(",")

        return int(size) if size else None

    @size.setter
    def size(self, value):
        self.value_text = ",".join((self.shape, str(value)))

    def test_valid(self, pipeline):
        if self.size is None:
            raise ValidationError(
                "Missing structuring element size. Please enter a positive integer.",
                self,
            )

        if self.size <= 0:
            raise ValidationError(
                "Structuring element size must be a positive integer. You provided {}.".format(
                    self.size
                ),
                self,
            )

        if pipeline.volumetric():
            if (
                self.shape in ["diamond", "disk", "square", "star"]
                and not self.__allow_planewise
            ):
                raise ValidationError(
                    "A 3 dimensional struturing element is required. You selected {}."
                    ' Please select one of "ball", "cube", or "octahedron".'.format(
                        self.shape
                    ),
                    self,
                )
        else:
            if self.shape in ["ball", "cube", "octahedron"]:
                raise ValidationError(
                    "A 2 dimensional structuring element is required. You selected {}."
                    ' Please select one of "diamond", "disk", "square", "star".'.format(
                        self.shape
                    ),
                    self,
                )



import logging
import pickle

from ._setting import Setting
from ._validation_error import ValidationError


LOGGER = logging.getLogger(__name__)

class ImagePlane(Setting):
    """A setting that specifies an image plane

    This setting lets the user pick an image plane. This might be an image
    file, a URL, or a plane from an image stack.

    The text setting has four fields, delimited by a space character (which
    is luckily disallowed in a URL:

    Field 1: URL

    Field 2: series (or None if a space is followed by another)

    Field 3: index (or None if blank)

    Field 4: channel (or None if blank)

    Field 5: Z (or None if blank)

    Field 6: T (or None if blank)

    """

    def __init__(self, text, *args, **kwargs):
        """Initialize the setting

        text - informative text to display to the left
        """
        super(ImagePlane, self).__init__(text, ImagePlane.build(""), *args, **kwargs)

    @staticmethod
    def build(plane):
        """Build the string representation of the setting

        plane - the ImagePlane object

        series - the series for a multi-series stack or None if the whole file

        index - the index of the frame for a multi-frame stack or None if
                the whole stack

        channel - the channel of an interlaced color image or None if all
                  channels
        """
        if not plane:
            return ""
        url = plane.url
        if " " in url:
            # Spaces are not legal characters in URLs, nevertheless, I try
            # to accommodate
            LOGGER.warning(
                "URLs should not contain spaces. %s is the offending URL" % url
            )
            url = url.replace(" ", "%20")
        return " ".join(
            [str(x) if x is not None else "" for x in (url, plane.series, plane.index, plane.channel, plane.z, plane.t)]
        )

    def get_plane(self):
        url = self.url
        if url is None:
            url = ""
        from ..pipeline import ImageFile
        from ..pipeline import ImagePlane as IPlane
        im_file = ImageFile(url)
        return IPlane(im_file, series=self.series, index=self.index, channel=self.channel, z=self.z, t=self.t)

    def __get_field(self, index):
        split = self.value_text.split(" ")
        if index >= len(split):
            return None
        f = split[index]
        if len(f) == 0:
            return None
        return f

    def __get_int_field(self, index):
        f = self.__get_field(index)
        if f is None:
            return f
        return int(f)

    @property
    def url(self):
        """The URL portion of the image plane descriptor"""
        uurl = self.__get_field(0)
        if uurl is not None:
            uurl = uurl
        return uurl

    @property
    def series(self):
        """The series portion of the image plane descriptor"""
        return self.__get_int_field(1)

    @property
    def index(self):
        """The index portion of the image plane descriptor"""
        return self.__get_int_field(2)

    @property
    def channel(self):
        """The channel portion of the image plane descriptor"""
        return self.__get_int_field(3)

    @property
    def z(self):
        """The channel portion of the image plane descriptor"""
        return self.__get_int_field(4)

    @property
    def t(self):
        """The channel portion of the image plane descriptor"""
        return self.__get_int_field(5)

    def test_valid(self, pipeline):
        if self.url is None:
            raise ValidationError(
                "This setting's URL is blank. Please select a valid image", self
            )



import logging
import os
import re
import shutil
import sys
import tempfile
import urllib.request
from urllib.parse import parse_qs, unquote, urlparse
from urllib.request import urlopen
import boto3
import numpy
import importlib.resources
import scipy.io
from google.cloud import storage
from google.cloud.exceptions import NotFound
from google.auth.exceptions import DefaultCredentialsError

from ..constants.image import (ALL_IMAGE_EXTENSIONS, FILE_SCHEME,
                               PASSTHROUGH_SCHEMES, SUPPORTED_MOVIE_EXTENSIONS)
from ..constants.measurement import FTR_WELL
from .measurement import is_well_column_token, is_well_row_token

LOGGER = logging.getLogger(__name__)

"""
This temporary directory will store cached files downloaded from the web.
It'll automatically delete on exit, and we create it early here so that it's
one of the last objects to get deleted - any opened files must close first.
"""
CP_TEMP_DIR = tempfile.TemporaryDirectory(prefix='cellprofiler_',)


def convert_image_to_objects(image):
    """Interpret an image as object indices

    image - a grayscale or color image, assumes zero == background

    returns - a similarly shaped integer array with zero representing background
              and other values representing the indices of the associated object.
    """
    assert isinstance(image, numpy.ndarray)
    if image.ndim == 2:
        unique_indices = numpy.unique(image.ravel())
        if len(unique_indices) * 2 > max(numpy.max(unique_indices), 254) and numpy.all(
            numpy.abs(numpy.round(unique_indices, 1) - unique_indices)
            <= numpy.finfo(float).eps
        ):
            # Heuristic: reinterpret only if sparse and roughly integer
            return numpy.round(image).astype(int)

        def sorting(x):
            return [x]

        def comparison(i0, i1):
            return image.ravel()[i0] != image.ravel()[i1]

    else:
        i, j = numpy.mgrid[0 : image.shape[0], 0 : image.shape[1]]

        def sorting(x):
            return [x[:, :, 2], x[:, :, 1], x[:, :, 0]]

        def comparison(i0, i1):
            return numpy.any(
                image[i.ravel()[i0], j.ravel()[i0], :]
                != image[i.ravel()[i1], j.ravel()[i1], :],
                1,
            )

    order = numpy.lexsort([x.ravel() for x in sorting(image)])
    different = numpy.hstack([[False], comparison(order[:-1], order[1:])])
    index = numpy.cumsum(different)
    image = numpy.zeros(image.shape[:2], index.dtype)
    image.ravel()[order] = index
    return image


def default_cpimage_name(index):
    # the usual suspects
    names = ["DNA", "Actin", "Protein"]
    if index < len(names):
        return names[index]
    return "Channel%d" % (index + 1)


def well_metadata_tokens(tokens):
    """Return the well row and well column tokens out of a set of metadata tokens"""

    well_row_token = None
    well_column_token = None
    for token in tokens:
        if is_well_row_token(token):
            well_row_token = token
        if is_well_column_token(token):
            well_column_token = token
    return well_row_token, well_column_token


def needs_well_metadata(tokens):
    """Return true if, based on a set of metadata tokens, we need a well token

    Check for a row and column token and the absence of the well token.
    """
    if FTR_WELL.lower() in [x.lower() for x in tokens]:
        return False
    well_row_token, well_column_token = well_metadata_tokens(tokens)
    return (well_row_token is not None) and (well_column_token is not None)


def is_image(filename):
    """Determine if a filename is a potential image file based on extension"""
    ext = os.path.splitext(filename)[1].lower()
    return ext in ALL_IMAGE_EXTENSIONS


def is_movie(filename):
    ext = os.path.splitext(filename)[1].lower()
    return ext in SUPPORTED_MOVIE_EXTENSIONS


def is_numpy_file(filename):
    return os.path.splitext(filename)[-1].lower() == ".npy"


def is_matlab_file(filename):
    return os.path.splitext(filename)[-1].lower() == ".mat"


def loadmat(path):
    imgdata = scipy.io.matlab.mio.loadmat(path, struct_as_record=True)
    img = imgdata["Image"]

    return img


def load_data_file(pathname_or_url, load_fn):
    ext = os.path.splitext(pathname_or_url)[-1].lower()

    if any([pathname_or_url.startswith(scheme) for scheme in PASSTHROUGH_SCHEMES]):
        url = generate_presigned_url(pathname_or_url)

        try:
            src = urllib.request.urlopen(url)
            fd, path = tempfile.mkstemp(suffix=ext)
            with os.fdopen(fd, mode="wb") as dest:
                shutil.copyfileobj(src, dest)
            img = load_fn(path)
        finally:
            try:
                src.close()
                os.remove(path)
            except NameError:
                pass

        return img

    return load_fn(pathname_or_url)


def is_file_url(url):
    return url.lower().startswith(FILE_SCHEME)


def is_image_extension(suffix):
    """Return True if the extension is one of those recongized by bioformats"""
    return suffix.lower() in ALL_IMAGE_EXTENSIONS


def crop_image(image, crop_mask, crop_internal=False):
    """Crop an image to the size of the nonzero portion of a crop mask"""
    i_histogram = crop_mask.sum(axis=1)
    i_cumsum = numpy.cumsum(i_histogram != 0)
    j_histogram = crop_mask.sum(axis=0)
    j_cumsum = numpy.cumsum(j_histogram != 0)
    if i_cumsum[-1] == 0:
        # The whole image is cropped away
        return numpy.zeros((0, 0), dtype=image.dtype)
    if crop_internal:
        #
        # Make up sequences of rows and columns to keep
        #
        i_keep = numpy.argwhere(i_histogram > 0)
        j_keep = numpy.argwhere(j_histogram > 0)
        #
        # Then slice the array by I, then by J to get what's not blank
        #
        return image[i_keep.flatten(), :][:, j_keep.flatten()].copy()
    else:
        #
        # The first non-blank row and column are where the cumsum is 1
        # The last are at the first where the cumsum is it's max (meaning
        # what came after was all zeros and added nothing)
        #
        i_first = numpy.argwhere(i_cumsum == 1)[0]
        i_last = numpy.argwhere(i_cumsum == i_cumsum.max())[0]
        i_end = i_last + 1
        j_first = numpy.argwhere(j_cumsum == 1)[0]
        j_last = numpy.argwhere(j_cumsum == j_cumsum.max())[0]
        j_end = j_last + 1

        if image.ndim == 3:
            return image[i_first[0] : i_end[0], j_first[0] : j_end[0], :].copy()

        return image[i_first[0] : i_end[0], j_first[0] : j_end[0]].copy()


def make_dictionary_key(key):
    """Make a dictionary into a stable key for another dictionary"""
    return ", ".join([":".join([str(y) for y in x]) for x in sorted(key.items())])


def image_resource(filename):
    try:
        if os.path.join("cellprofiler", "docs", "frontend") in os.path.abspath(os.curdir).lower():
            # We're probably trying to buld the manual
            thepath = os.path.relpath(
                os.path.abspath(
                    importlib.resources.files("cellprofiler").joinpath(
                        "..", "images", filename
                    )
                ),
                os.path.abspath(os.curdir),
            )
        else:
            if hasattr(sys, "frozen"):
                # Use relative paths if we're frozen.
                thepath = os.path.relpath(
                    importlib.resources.files("cellprofiler").joinpath(
                        "data", "images", filename
                    )
                )
            else:
                thepath = os.path.abspath(
                    importlib.resources.files("cellprofiler").joinpath(
                        "data", "images", filename
                    )
                )
        return thepath.replace("\\", "/")
    except ModuleNotFoundError:
        # CellProfiler is not installed so the assets are missing.
        # In theory an icon should never be called without the GUI anyway
        print("CellProfiler image assets were not found")
    return ""


def generate_presigned_url(url):
    """
    Generate a presigned URL, if necessary (e.g., s3).

    :param url: An unsigned URL.
    :return: The presigned URL.
    """
    if url.startswith("s3"):
        client = boto3.client("s3")

        bucket_name, filename = (
            re.compile("s3://([\w\d\-.]+)/(.*)").search(url).groups()
        )

        url = client.generate_presigned_url(
            "get_object",
            Params={"Bucket": bucket_name, "Key": filename.replace("+", " ")},
        )

    return url


def url_to_modpath(url):
    import urllib.parse

    from .hdf5_dict import HDF5FileList

    if not url.lower().startswith("file:"):
        schema, rest = HDF5FileList.split_url(url)
        return [schema] + rest[0:1] + [unquote(part) for part in rest[1:]]
    path = urllib.request.url2pathname(url[5:])
    parts = []
    while True:
        new_path, part = os.path.split(path)
        if len(new_path) == 0 or len(part) == 0:
            parts.insert(0, path)
            break
        parts.insert(0, part)
        path = new_path
    return parts

def _handle_gcs_url(netloc, ext, urlpath):
    try:
        # Create client to access Google Cloud Storage.
        client = storage.Client()
        # Get bucket from bucket name parsed from URL.
        bucket = client.get_bucket(netloc)
        # Create a blob object from the given filepath.
        urlpath = urlpath[1:] if urlpath.startswith("/") else urlpath
        blob = bucket.blob(urlpath)
        # Provision a temporary file to which to download the blob.
        dest_file = tempfile.NamedTemporaryFile(suffix=ext, delete=False, dir=CP_TEMP_DIR.name)
        blob.download_to_filename(dest_file.name)
        return dest_file.name
    except DefaultCredentialsError as e:
        LOGGER.error(f"Unable to download Google Cloud Storage image to temp file: {e.args[0]}")
        return None
    except NotFound as e:
        # 404 error
        LOGGER.error(f"Unable to download Google Cloud Storage image to temp file. {e}")
        dest_file.close()
        return None
    except Exception as e:
        LOGGER.error(f"Unable to download Google Cloud Storage image to temp file. {e}")
        if dest_file:
            dest_file.close()
        return None

def download_to_temp_file(url):
    global CP_TEMP_DIR
    parsed = urlparse(url)
    scheme = parsed.scheme
    netloc = parsed.netloc
    path = parsed.path
    queries = parse_qs(parsed.query)
    if 'name' in queries:
        path = queries['name'][0]
    ext = os.path.splitext(path)[-1]

    # Handle Google Cloud Storage URLs.
    if scheme == 'gs':
        return _handle_gcs_url(netloc, ext, path)
    else:
        # Handle Amazon Web Services URLs.
        if scheme == 's3':
            client = boto3.client('s3')
            bucket_name, key = re.compile('s3://([\w\d\-\.]+)/(.*)').search(url).groups()
            # Get pre-signed URL.
            url = client.generate_presigned_url(
                'get_object',
                Params={'Bucket': bucket_name, 'Key': key.replace("+", " ")}
            )
        # Download object from URL.
        src = urlopen(url)
        dest_file = tempfile.NamedTemporaryFile(suffix=ext, delete=False, dir=CP_TEMP_DIR.name)
        try:
            shutil.copyfileobj(src, dest_file)
        except Exception as e:
            LOGGER.error(f"Unable to download image to temp file. {e}")
            return None
        finally:
            dest_file.close()
        return dest_file.name




class RelationshipKey:
    def __init__(self, module_number, relationship, object_name1, object_name2):
        self.module_number = module_number
        self.relationship = relationship
        self.object_name1 = object_name1
        self.object_name2 = object_name2



from ._change_setting_event import ChangeSettingEvent


class BeforeChangeSettingEvent(ChangeSettingEvent):
    """Indicates that a setting is about to change, allows a listener to cancel the change

    """

    def __init__(self, old_value, new_value):
        ChangeSettingEvent.__init__(self, old_value, new_value)
        self.__allow_change = True
        self.__cancel_reason = None

    def cancel_change(self, reason=None):
        self.__allow_change = False
        self.__cancel_reason = reason

    def allow_change(self):
        return self.__allow_change

    def get_cancel_reason(self):
        return self.__cancel_reason

    cancel_reason = property(get_cancel_reason)



from ._setting import Setting


class PathListDisplay(Setting):
    """This setting's only purpose is to signal that the path list should be shown

    Set self.using_filter to True if the module knows that the path list will
    be filtered or if the module doesn't know. Set it to False if the module
    knows the path list won't be filtered.
    """

    def __init__(self):
        super(self.__class__, self).__init__("", value="")
        self.using_filter = True



import os
from .....utilities.pathname import url2pathname

from .._file_image import FileImage


class URLImage(FileImage):
    """Reference an image via a URL"""

    def __init__(
        self,
        name,
        url,
        rescale_range=None,
        metadata_rescale=False,
        series=None,
        index=None,
        channel=None,
        volume=False,
        spacing=None,
        z=None,
        t=None
    ):
        if url.lower().startswith("file:"):
            path = url2pathname(url)
            pathname, filename = os.path.split(path)
        else:
            pathname = ""
            filename = url
        super(URLImage, self).__init__(
            name,
            pathname,
            filename,
            rescale_range=rescale_range,
            metadata_rescale=metadata_rescale,
            series=series,
            index=index,
            channel=channel,
            volume=volume,
            spacing=spacing,
            z=z,
            t=t
        )
        self.url = url

    def get_url(self):
        if self.cache_file():
            return super(URLImage, self).get_url()
        return self.url



from cellprofiler_library.functions.segmentation import convert_dense_to_sparse
from cellprofiler_library.functions.segmentation import convert_sparse_to_dense
from cellprofiler_library.functions.segmentation import indices_from_dense
from cellprofiler_library.functions.segmentation import dense_shape_from_sparse
from cellprofiler_library.functions.segmentation import _validate_dense
from cellprofiler_library.functions.segmentation import _validate_sparse
from cellprofiler_library.functions.segmentation import _validate_dense_shape


class Segmentation:
    """A segmentation of a space into labeled objects

    Supports overlapping objects and cacheing. Retrieval can be as a
    single plane (legacy), as multiple planes and as sparse ijv.
    """

    def __init__(self, dense=None, sparse=None, shape=None):
        """Initialize the segmentation with either a dense or sparse labeling

        dense - A 6-d array composed of one or more 5-d integer labelings of
        each hyper-voxel. The dimension order is labeling, c, t, z, y, x.
        Typically, a 2-D non-overlapping segmentation has dimensions of
        1, 1, 1, 1, y, x.

        sparse - a labeling stored in a record data type with each column
                 having a name of "c", "t", "z", "y", "x" or "label".
                 The "label" column is the object number, starting with 1.
                 When "c", "t", "z" are absent, this is the interoperable with
                 an ijv labeling of the pixels, or in the COOrdinate format
                 (as in scipy.sparse.coo_matrix).

        shape - the 5-D shape of the imaging site if sparse. If this is absent
                the shape is inferred from the given coordinates of the sparse
                labeling, which is a size capable of containing an equivilent
                dense representation, but may not be exactly equal to the
                original shape of the imaging site.
        """
        if dense is not None:
            _validate_dense(dense)
        if sparse is not None:
            _validate_sparse(sparse)
        if shape is not None:
            _validate_dense_shape(shape)

        self.__dense = dense
        self.__sparse = sparse
        if shape is not None:
            self.__shape = shape
            self.__explicit_shape = True
        else:
            self.__shape = None
            self.__explicit_shape = False

        if dense is not None:
            self.__indices = indices_from_dense(dense)

    @property
    def shape(self):
        """Get or estimate the shape of the segmentation matrix
        This is the dense shape, ('c', 't', 'z', 'y', 'x')

        Order of precedence:
        Shape supplied in the constructor
        Shape of the dense representation
        maximum extent of the sparse representation + 1
        """
        if self.__shape is not None:
            return self.__shape
        if self.has_dense():
            self.__shape = self.get_dense()[0].shape[1:]
        else:
            sparse = self.sparse
            if len(sparse) == 0:
                self.__shape = (1, 1, 1, 1, 1)
            else:
                self.__shape = dense_shape_from_sparse(sparse, validate=False)
        return self.__shape

    @shape.setter
    def shape(self, shape):
        """Set the shape of the segmentation array

        shape - the 5D shape of the array

        This fixes the shape of the 5D array for sparse representations
        """
        self.__shape = shape
        self.__explicit_shape = True

    def has_dense(self):
        return self.__dense is not None

    def has_sparse(self):
        return self.__sparse is not None

    def has_shape(self):
        if self.__explicit_shape:
            return True

        return self.has_dense()

    @property
    def sparse(self):
        """Get the sparse representation of the segmentation

        returns a Numpy record array where every row represents
        the labeling of a pixel. The dtype record names are taken from
        HDF5ObjectSet.AXIS_[C,T,Z,Y,X] and AXIS_LABELS for the object
        numbers.
        """
        if self.__sparse is not None:
            return self.__sparse

        if not self.has_dense():
            raise ValueError("Can't find object dense segmentation.")

        return self.__convert_dense_to_sparse()

    def get_dense(self):
        """Get the dense representation of the segmentation

        return the segmentation as a 6-D array and a sequence of arrays of the
        object numbers in each 5-D hyperplane of the segmentation. The first
        axis of the segmentation allows us to assign multiple labels to
        individual pixels. Given a 5-D algorithm, the code typically iterates
        over the first axis:

        for labels in self.get_dense():
            # do something

        The remaining axes are in the order, c, t, z, y and x
        """
        if self.__dense is not None:
            return self.__dense, self.__indices

        if not self.has_sparse():
            raise ValueError("Can't find object sparse segmentation.")

        return self.__convert_sparse_to_dense()

    def __convert_dense_to_sparse(self):
        dense, _ = self.get_dense()
        sparse = convert_dense_to_sparse(dense, validate=False)
        self.__sparse = sparse
        return sparse

    def __set_dense(self, dense, indices=None):
        self.__dense = dense

        if indices is not None:
            self.__indices = indices
        else:
            self.__indices = indices_from_dense(dense)

        return dense, self.__indices

    def __convert_sparse_to_dense(self):
        dense, indices = convert_sparse_to_dense(
            self.sparse, self.shape, validate=False)

        return self.__set_dense(dense, indices)



from ._do_something import DoSomething


class RemoveSettingButton(DoSomething):
    """A button whose only purpose is to remove something from a list."""

    def __init__(self, text, label, list, entry, **kwargs):
        super(RemoveSettingButton, self).__init__(
            text, label, lambda: list.remove(entry), **kwargs
        )



from .._range import Range


class IntegerRange(Range):
    """A setting that allows only integer input between two constrained values
    """

    valid_format_text = "%s must be all digits"

    def __init__(self, text, value=(0, 1), minval=None, maxval=None, *args, **kwargs):
        """Initialize an integer range
        text  - helpful text to be displayed to the user
        value - initial default value, a two-tuple as minimum and maximum
        minval - the minimum acceptable value of either
        maxval - the maximum acceptable value of either
        """
        super(IntegerRange, self).__init__(
            text, "%d,%d" % value, minval, maxval, *args, **kwargs
        )

    def str_to_value(self, value_str):
        return int(value_str)

    def value_to_str(self, value):
        return "%d" % value



from ._text import Text


class Filename(Text):
    """A setting that displays a file name

    optional arguments -
       get_directory_fn is a function that gets the initial directory
           for the browse button
       set_directory_fn is a function that sets the directory after browsing
       browse_msg - message at top of file browser
       exts - a list of tuples where the first is the user-displayed text
       and the second is the file filter for an extension, like
       [("Pipeline (*.cp)","*.cp")]
       mode - Controls whether a file-open or file-save dialog is displayed
              when the user browses.
              FilenameText.MODE_OPEN - open a file that must exist
              FilenameText.MODE_APPEND - open a file for modification or
                    create a new file (using the Save dialog)
              FilenameText.MODE_OVERWRITE - create a new file and warn the
                    user if the file exists and will be overwritten.
    """

    MODE_OPEN = "Open"
    MODE_APPEND = "Append"
    MODE_OVERWRITE = "Overwrite"

    def __init__(self, text, value, *args, **kwargs):
        kwargs = kwargs.copy()
        self.get_directory_fn = kwargs.pop("get_directory_fn", None)
        self.set_directory_fn = kwargs.pop("set_directory_fn", None)
        self.browse_msg = kwargs.pop("browse_msg", "Choose a file")
        self.exts = kwargs.pop("exts", None)
        self.mode = kwargs.pop("mode", self.MODE_OPEN)
        super(Filename, self).__init__(text, value, *args, **kwargs)
        self.browsable = True

    def set_browsable(self, val):
        self.browsable = val



from ...utilities.zmq.communicable.request import AnalysisRequest


class DisplayPostGroup(AnalysisRequest):
    """Request a post-group display

    This is a message sent to the UI from the analysis worker"""

    def __init__(self, analysis_id, module_num, display_data, image_set_number):
        AnalysisRequest.__init__(
            self,
            analysis_id,
            module_num=module_num,
            image_set_number=image_set_number,
            display_data=display_data,
        )



from .._alphanumeric import Alphanumeric


class Name(Alphanumeric):
    """
    A setting that provides a named object
    """

    def __init__(self, text, group, value="Do not use", *args, **kwargs):
        self.__provided_attributes = {"group": group}
        kwargs = kwargs.copy()
        if "provided_attributes" in kwargs:
            self.__provided_attributes.update(kwargs["provided_attributes"])
            del kwargs["provided_attributes"]
        kwargs["first_must_be_alpha"] = True
        super(Name, self).__init__(text, value, *args, **kwargs)

    def get_group(self):
        """
        This setting provides a name to this group

        Returns a group name, e.g., imagegroup or objectgroup
        """
        return self.__provided_attributes["group"]

    group = property(get_group)

    @property
    def provided_attributes(self):
        """
        Return the dictionary of attributes of this provider

        These are things like the group ("objectgroup" for instance) and
        hints about the thing itself, such as that it is an image
        that was loaded from  a file.
        """
        return self.__provided_attributes



import collections
import io
import logging
import os
import queue
import subprocess
import sys
import tempfile
import threading
import re

import numpy
import psutil

from .event import Finished
from .event import Paused
from .event import Progress
from .event import Resumed
from .event import Started
from . import reply as anareply
from . import request as anarequest
from ..image import ImageSetList
from ..measurement import Measurements
from ..utilities.measurement import load_measurements_from_buffer
from ..pipeline import dump
from ..preferences import get_plugin_directory
from ..preferences import get_conserve_memory
from ..preferences import get_temporary_directory
from ..preferences import get_always_continue
from ..preferences import preferences_as_dict
from ..utilities.analysis import close_all_on_exec, start_daemon_thread
from ..utilities.analysis import find_analysis_worker_source
from ..utilities.analysis import find_python
from ..utilities.analysis import find_worker_env
from ..utilities.zmq import start_boundary, Boundary
from ..utilities.zmq.communicable.reply import Reply
from ..workspace import Workspace


LOGGER = logging.getLogger(__name__)

class Runner:
    """The Runner manages two threads (per instance) and all of the
    workers (per class, i.e., singletons).

    The two threads run interface() and jobserver(), below.

    interface() is responsible grouping jobs for dispatch, tracking job
    progress, integrating measurements returned from workers.

    jobserver() is a lightweight thread that serves jobs and receives their
    requests, acting as a switchboard between workers, interface(), and
    whatever event_listener is present (via post_event()).

    workers are stored in AnalysisRunner.workers[], and are separate processes.
    They are expected to exit if their stdin() closes, e.g., if the parent
    process goes away.

    interface() and jobserver() communicate via Queues and using condition
    variables to get each other's attention.  zmqrequest is used to communicate
    between jobserver() and workers[].
    """

    # measurement status
    STATUS = "ProcessingStatus"
    STATUS_UNPROCESSED = "Unprocessed"
    STATUS_IN_PROCESS = "InProcess"
    STATUS_FINISHED_WAITING = "FinishedWaitingMeasurements"
    STATUS_DONE = "Done"
    STATUSES = [
        STATUS_UNPROCESSED,
        STATUS_IN_PROCESS,
        STATUS_FINISHED_WAITING,
        STATUS_DONE,
    ]

    def __init__(
        self, analysis_id, pipeline, initial_measurements_buf, event_listener,
    ):
        self.initial_measurements_buf = initial_measurements_buf

        self.analysis_id = analysis_id
        self.pipeline = pipeline.copy(preserve_module_state=True)
        self.event_listener = event_listener

        self.interface_work_cv = threading.Condition()
        self.jobserver_work_cv = threading.Condition()
        self.paused = False
        self.cancelled = False

        self.work_queue = queue.Queue()
        self.in_process_queue = queue.Queue()
        self.finished_queue = queue.Queue()

        self.workers = []

        # We use a queue size of 10 because we keep measurements in memory (as
        # their HDF5 file contents) until they get merged into the full
        # measurements set.  If at some point, this size is too limiting, we
        # should have jobserver() call load_measurements_from_buffer() rather
        # than interface() doing so.  Currently, passing measurements in this
        # way seems like it might be buggy:
        # https://github.com/h5py/h5py/issues/244
        self.received_measurements_queue = queue.Queue(maxsize=10)

        self.shared_dicts = None

        self.boundary = None
        self.interface_thread = None
        self.jobserver_thread = None

    # External control interfaces
    def start(self, num_workers=None, overwrite=True):
        """start the analysis run

        num_workers - # of workers to run, default = # of cores
        overwrite - if True, overwrite existing image set measurements, False
                    try to reuse them.
        """

        # Although it would be nice to reuse the worker pool, I'm not entirely
        # sure they recover correctly from the user cancelling an analysis
        # (e.g., during an interaction request).  This should be handled by
        # zmqRequest.cancel_analysis, but just in case, we stop the workers and
        # restart them.  Note that this creates a new announce port, so we
        # don't have to worry about old workers taking a job before noticing
        # that their stdin has closed.
        self.stop_workers()

        self.boundary = Boundary("tcp://127.0.0.1")

        start_signal = threading.Semaphore(0)
        self.interface_thread = start_daemon_thread(
            target=self.interface,
            args=(start_signal,),
            kwargs=dict(overwrite=overwrite),
            name="AnalysisRunner.interface thread",
        )
        #
        # Wait for signal on interface started.
        #
        start_signal.acquire()
        self.jobserver_thread = start_daemon_thread(
            target=self.jobserver,
            args=(start_signal, ),
            name="AnalysisRunner.jobserver thread",
        )
        #
        # Wait for signal on jobserver started.
        #
        start_signal.acquire()
        # start worker pool via class method (below)
        self.start_workers(num_workers)

    def check(self):
        return (
            (self.interface_thread is not None)
            and (self.jobserver_thread is not None)
            and self.interface_thread.is_alive()
            and self.jobserver_thread.is_alive()
        )

    def notify_threads(self):
        with self.interface_work_cv:
            self.interface_work_cv.notify()
        with self.jobserver_work_cv:
            self.jobserver_work_cv.notify()

    def cancel(self):
        """cancel the analysis run"""
        LOGGER.debug("Stopping workers")
        self.stop_workers()
        LOGGER.debug("Canceling run")
        self.cancelled = True
        self.paused = False
        self.notify_threads()
        LOGGER.debug("Waiting on interface thread")
        self.interface_thread.join()
        LOGGER.debug("Waiting on jobserver thread")
        self.jobserver_thread.join()
        self.interface_thread = None
        self.jobserver_thread = None
        self.work_queue = queue.Queue()
        self.in_process_queue = queue.Queue()
        self.finished_queue = queue.Queue()
        LOGGER.debug("Cancel complete")

    def pause(self):
        """pause the analysis run"""
        self.paused = True
        self.notify_threads()

    def resume(self):
        """resume a paused analysis run"""
        self.paused = False
        self.notify_threads()

    # event posting
    def post_event(self, evt):
        self.event_listener(evt)

    def post_run_display_handler(self, workspace, module):
        event = anarequest.DisplayPostRun(module.module_num, workspace.display_data)
        self.event_listener(event)

    # XXX - catch and deal with exceptions in interface() and jobserver() threads
    def interface(
        self, start_signal, image_set_start=1, image_set_end=None, overwrite=True
    ):
        """Top-half thread for running an analysis.  Sets up grouping for jobs,
        deals with returned measurements, reports status periodically.

        start_signal- signal this semaphore when jobs are ready.
        image_set_start - beginning image set number to process
        image_set_end - last image set number to process
        overwrite - whether to recompute imagesets that already have data in initial_measurements.
        """
        posted_analysis_started = False
        acknowledged_thread_start = False
        measurements = None
        workspace = None
        try:
            # listen for pipeline events, and pass them upstream
            self.pipeline.add_listener(lambda pipe, evt: self.post_event(evt))

            initial_measurements = None
            # Make a temporary measurements file.
            fd, filename = tempfile.mkstemp(".h5", dir=get_temporary_directory())
            try:
                fd = os.fdopen(fd, "wb")
                fd.write(self.initial_measurements_buf)
                fd.close()
                initial_measurements = Measurements(filename=filename, mode="r")
                measurements = Measurements(
                    image_set_start=None, copy=initial_measurements, mode="a"
                )
            finally:
                if initial_measurements is not None:
                    initial_measurements.close()
                os.unlink(filename)

            # The shared dicts are needed in jobserver()
            self.shared_dicts = [m.get_dictionary() for m in self.pipeline.modules()]
            workspace = Workspace(
                self.pipeline, None, None, None, measurements, ImageSetList(),
            )

            if image_set_end is None:
                image_set_end = measurements.get_image_numbers()[-1]
            image_sets_to_process = list(
                filter(
                    lambda x: image_set_start <= x <= image_set_end,
                    measurements.get_image_numbers(),
                )
            )

            self.post_event(Started())
            posted_analysis_started = True

            # reset the status of every image set that needs to be processed
            has_groups = measurements.has_groups()
            if self.pipeline.requires_aggregation():
                overwrite = True
            if has_groups and not overwrite:
                if not measurements.has_feature("Image", self.STATUS):
                    overwrite = True
                else:
                    group_status = {}
                    for image_number in measurements.get_image_numbers():
                        group_number = measurements[
                            "Image", "Group_Number", image_number,
                        ]
                        status = measurements[
                            "Image", self.STATUS, image_number,
                        ]
                        if status != self.STATUS_DONE:
                            group_status[group_number] = self.STATUS_UNPROCESSED
                        elif group_number not in group_status:
                            group_status[group_number] = self.STATUS_DONE

            new_image_sets_to_process = []
            for image_set_number in image_sets_to_process:
                needs_reset = False
                if (
                    overwrite
                    or (
                        not measurements.has_measurements(
                            "Image", self.STATUS, image_set_number,
                        )
                    )
                    or (
                        measurements["Image", self.STATUS, image_set_number,]
                        != self.STATUS_DONE
                    )
                ):
                    needs_reset = True
                elif has_groups:
                    group_number = measurements[
                        "Image", "Group_Number", image_set_number,
                    ]
                    if group_status[group_number] != self.STATUS_DONE:
                        needs_reset = True
                if needs_reset:
                    new_image_sets_to_process.append(image_set_number)
            measurements["Image", self.STATUS, new_image_sets_to_process] = \
                [self.STATUS_UNPROCESSED] * len(new_image_sets_to_process)
            image_sets_to_process = new_image_sets_to_process

            # Find image groups.  These are written into measurements prior to
            # analysis.  Groups are processed as a single job.
            if has_groups or self.pipeline.requires_aggregation():
                worker_runs_post_group = True
                job_groups = {}
                for image_set_number in image_sets_to_process:
                    group_number = measurements[
                        "Image", "Group_Number", image_set_number,
                    ]
                    group_index = measurements[
                        "Image", "Group_Index", image_set_number,
                    ]
                    job_groups[group_number] = job_groups.get(group_number, []) + [
                        (group_index, image_set_number)
                    ]
                job_groups = [
                    [isn for _, isn in sorted(job_groups[group_number])]
                    for group_number in sorted(job_groups)
                ]
            else:
                worker_runs_post_group = False  # prepare_group will be run in worker, but post_group is below.
                job_groups = [
                    [image_set_number] for image_set_number in image_sets_to_process
                ]

            # XXX - check that any constructed groups are complete, i.e.,
            # image_set_start and image_set_end shouldn't carve them up.

            if not worker_runs_post_group:
                # put the first job in the queue, then wait for the first image to
                # finish (see the check of self.finish_queue below) to post the rest.
                # This ensures that any shared data from the first imageset is
                # available to later imagesets.
                self.work_queue.put((job_groups[0], worker_runs_post_group, True))
                waiting_for_first_imageset = True
                del job_groups[0]
            else:
                waiting_for_first_imageset = False
                for job in job_groups:
                    self.work_queue.put((job, worker_runs_post_group, False))
                job_groups = []
            start_signal.release()
            acknowledged_thread_start = True

            # We loop until every image is completed, or an outside event breaks the loop.
            while not self.cancelled:

                # gather measurements
                while not self.received_measurements_queue.empty():
                    image_numbers, buf = self.received_measurements_queue.get()
                    image_numbers = [int(i) for i in image_numbers]
                    recd_measurements = load_measurements_from_buffer(buf)
                    self.copy_recieved_measurements(
                        recd_measurements, measurements, image_numbers
                    )
                    recd_measurements.close()
                    del recd_measurements

                # check for jobs in progress
                while not self.in_process_queue.empty():
                    image_set_numbers = self.in_process_queue.get()
                    for image_set_number in image_set_numbers:
                        measurements[
                            "Image", self.STATUS, int(image_set_number),
                        ] = self.STATUS_IN_PROCESS

                # check for finished jobs that haven't returned measurements, yet
                while not self.finished_queue.empty():
                    finished_req = self.finished_queue.get()
                    measurements[
                        "Image", self.STATUS, int(finished_req.image_set_number),
                    ] = self.STATUS_FINISHED_WAITING
                    if waiting_for_first_imageset:
                        assert isinstance(
                            finished_req, anareply.ImageSetSuccessWithDictionary,
                        )
                        self.shared_dicts = finished_req.shared_dicts
                        waiting_for_first_imageset = False
                        assert len(self.shared_dicts) == len(self.pipeline.modules())
                        for worker_dict, module in zip(self.shared_dicts, self.pipeline.modules()):
                            # Apply those imported states to the master pipeline
                            module.get_dictionary().update(worker_dict)
                        # if we had jobs waiting for the first image set to finish,
                        # queue them now that the shared state is available.
                        for job in job_groups:
                            self.work_queue.put((job, worker_runs_post_group, False))
                    finished_req.reply(anareply.Ack())

                # check progress and report
                counts = collections.Counter(
                    measurements["Image", self.STATUS, image_set_number,]
                    for image_set_number in image_sets_to_process
                )
                self.post_event(Progress(counts))

                # Are we finished?
                if counts[self.STATUS_DONE] == len(image_sets_to_process):
                    last_image_number = measurements.get_image_numbers()[-1]
                    measurements.image_set_number = last_image_number
                    if not worker_runs_post_group:
                        self.pipeline.post_group(workspace, {})

                    workspace = Workspace(
                        self.pipeline, None, None, None, measurements, None, None
                    )
                    workspace.post_run_display_handler = self.post_run_display_handler
                    self.pipeline.post_run(workspace)
                    break

                measurements.flush()
                # not done, wait for more work
                with self.interface_work_cv:
                    while self.paused or (
                        (not self.cancelled)
                        and self.in_process_queue.empty()
                        and self.finished_queue.empty()
                        and self.received_measurements_queue.empty()
                    ):
                        self.interface_work_cv.wait()  # wait for a change of status or work to arrive
        except Exception as e:
            print(e)
        finally:
            # Note - the measurements file is owned by the queue consumer
            #        after this post_event.
            #
            if not acknowledged_thread_start:
                start_signal.release()
            if posted_analysis_started:
                was_cancelled = self.cancelled
                self.post_event(Finished(measurements, was_cancelled))
        self.analysis_id = False  # this will cause the jobserver thread to exit

    def copy_recieved_measurements(
        self, recd_measurements, measurements, image_numbers
    ):
        """Copy the received measurements to the local process' measurements

        recd_measurements - measurements received from worker

        measurements - local measurements = destination for copy

        image_numbers - image numbers processed by worker
        """
        measurements.copy_relationships(recd_measurements)
        for o in recd_measurements.get_object_names():
            if o == "Experiment":
                continue  # Written during prepare_run / post_run
            elif o == "Image":
                # Some have been previously written. It's worth the time
                # to check values and only write changes
                for feature in recd_measurements.get_feature_names(o):
                    if not measurements.has_feature("Image", feature):
                        f_image_numbers = image_numbers
                    else:
                        local_values = measurements["Image", feature, image_numbers]
                        remote_values = recd_measurements[
                            "Image", feature, image_numbers
                        ]
                        f_image_numbers = [
                            i
                            for i, lv, rv in zip(
                                image_numbers, local_values, remote_values
                            )
                            if (
                                numpy.any(rv != lv)
                                if isinstance(lv, numpy.ndarray)
                                else lv != rv
                            )
                        ]
                    if len(f_image_numbers) > 0:
                        measurements[o, feature, f_image_numbers] = recd_measurements[
                            o, feature, f_image_numbers
                        ]
            else:
                for feature in recd_measurements.get_feature_names(o):
                    measurements[o, feature, image_numbers] = recd_measurements[
                        o, feature, image_numbers
                    ]
        for image_set_number in image_numbers:
            measurements["Image", self.STATUS, image_set_number] = self.STATUS_DONE

    def jobserver(self, start_signal):
        # this server subthread should be very lightweight, as it has to handle
        # all the requests from workers, of which there might be several.

        # start the zmqrequest Boundary
        request_queue = queue.Queue()
        self.boundary.register_analysis(self.analysis_id, request_queue)
        #
        # The boundary is announcing our analysis at this point. Workers
        # will get announcements if they connect.
        #
        start_signal.release()

        # XXX - is this just to keep from posting another AnalysisPaused event?
        # If so, probably better to simplify the code and keep sending them
        # (should be only one per second).
        i_was_paused_before = False

        # start serving work until the analysis is done (or changed)
        while not self.cancelled:

            with self.jobserver_work_cv:
                if self.paused and not i_was_paused_before:
                    self.post_event(Paused())
                    i_was_paused_before = True
                if self.paused or request_queue.empty():
                    self.jobserver_work_cv.wait(
                        1
                    )  # we timeout in order to keep announcing ourselves.
                    continue  # back to while... check that we're still running

            if i_was_paused_before:
                self.post_event(Resumed())
                i_was_paused_before = False

            try:
                req = request_queue.get(timeout=0.25)
            except queue.Empty:
                continue

            if isinstance(req, anarequest.PipelinePreferences):
                LOGGER.debug("Received pipeline preferences request")
                req.reply(
                    Reply(
                        pipeline_blob=numpy.array(self.pipeline_as_string()),
                        preferences=preferences_as_dict(),
                    )
                )
                LOGGER.debug("Replied to pipeline preferences request")
            elif isinstance(req, anarequest.InitialMeasurements):
                LOGGER.debug("Received initial measurements request")
                req.reply(Reply(buf=self.initial_measurements_buf))
                LOGGER.debug("Replied to initial measurements request")
            elif isinstance(req, anarequest.Work):
                if not self.work_queue.empty():
                    LOGGER.debug("Received work request")
                    (
                        job,
                        worker_runs_post_group,
                        wants_dictionary,
                    ) = self.work_queue.get()
                    req.reply(
                        anareply.Work(
                            image_set_numbers=job,
                            worker_runs_post_group=worker_runs_post_group,
                            wants_dictionary=wants_dictionary,
                        )
                    )
                    self.queue_dispatched_job(job)
                    LOGGER.debug(
                        "Dispatched job: image sets=%s"
                        % ",".join([str(i) for i in job])
                    )
                else:
                    # there may be no work available, currently, but there
                    # may be some later.
                    req.reply(anareply.NoWork())
            elif isinstance(req, anareply.ImageSetSuccess):
                # interface() is responsible for replying, to allow it to
                # request the shared_state dictionary if needed.
                LOGGER.debug("Received ImageSetSuccess")
                self.queue_imageset_finished(req)
                LOGGER.debug("Enqueued ImageSetSuccess")
            elif isinstance(req, anarequest.SharedDictionary):
                LOGGER.debug("Received shared dictionary request")
                req.reply(anareply.SharedDictionary(dictionaries=self.shared_dicts))
                LOGGER.debug("Sent shared dictionary reply")
            elif isinstance(req, anarequest.MeasurementsReport):
                LOGGER.debug("Received measurements report")
                self.queue_received_measurements(req.image_set_numbers, req.buf)
                req.reply(anareply.Ack())
                LOGGER.debug("Acknowledged measurements report")
            elif isinstance(req, anarequest.AnalysisCancel):
                # Signal the interface that we are cancelling
                LOGGER.debug("Received analysis worker cancel request")
                with self.interface_work_cv:
                    self.cancelled = True
                    self.interface_work_cv.notify()
                req.reply(anareply.Ack())
            elif isinstance(
                req,
                (
                    anarequest.Interaction,
                    anarequest.Display,
                    anarequest.DisplayPostGroup,
                    anarequest.ExceptionReport,
                    anarequest.DebugWaiting,
                    anarequest.DebugComplete,
                    anarequest.OmeroLogin,
                ),
            ):
                LOGGER.debug("Enqueueing interactive request")
                # bump upward
                self.post_event(req)
                LOGGER.debug("Interactive request enqueued")
            else:
                msg = "Unknown request from worker: %s of type %s" % (req, type(req))
                LOGGER.error(msg)
                raise ValueError(msg)

        # stop the ZMQ-boundary thread - will also deal with any requests waiting on replies
        if self.boundary is not None:
            self.boundary.cancel(self.analysis_id)

    def queue_job(self, image_set_number):
        self.work_queue.put(image_set_number)

    def queue_dispatched_job(self, job):
        self.in_process_queue.put(job)
        # notify interface thread
        with self.interface_work_cv:
            self.interface_work_cv.notify()

    def queue_imageset_finished(self, finished_req):
        self.finished_queue.put(finished_req)
        # notify interface thread
        with self.interface_work_cv:
            self.interface_work_cv.notify()

    def queue_received_measurements(self, image_set_numbers, measurements):
        did_block = False
        if self.received_measurements_queue.full():
            did_block = True
            LOGGER.debug("Received measurements queue is full, blocking until space is available")
        self.received_measurements_queue.put((image_set_numbers, measurements))
        if did_block:
            LOGGER.debug("Received measurements queue no longer blocked")
        # notify interface thread
        with self.interface_work_cv:
            self.interface_work_cv.notify()

    def pipeline_as_string(self):
        s = io.StringIO()
        dump(self.pipeline, s, version=5, save_image_plane_details=False)
        return s.getvalue()

    def start_workers(self, num=None):
        if self.workers:
            return

        if num is None:
            num = psutil.cpu_count(logical=False)

        boundary = self.boundary

        LOGGER.info("Starting workers on address %s" % boundary.request_address)

        close_fds = False

        aw_args = [
            "--analysis-id",
            self.analysis_id,
            "--work-server",
            boundary.request_address,
            "--notify-server",
            boundary.keepalive_address,
            "--plugins-directory",
            get_plugin_directory(),
            "--conserve-memory",
            str(get_conserve_memory()),
            "--always-continue",
            str(get_always_continue()),
            "--log-level",
            str(logging.root.level)
        ]

        # start workers
        for idx in range(num):
            if sys.platform == "darwin":
                close_all_on_exec()

            if hasattr(sys, "frozen"):
                if sys.platform == "darwin":
                    executable = os.path.join(os.path.dirname(sys.executable), "cp")
                    args = [executable] + aw_args
                elif sys.platform.startswith("linux"):
                    aw_path = os.path.join(os.path.dirname(__file__), "__init__.py")
                    args = [sys.executable, aw_path] + aw_args
                else:
                    args = [sys.executable] + aw_args

                worker = subprocess.Popen(
                    args,
                    env=find_worker_env(idx),
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    close_fds=close_fds,
                )
            else:
                worker = subprocess.Popen(
                    [find_python(), "-u", find_analysis_worker_source(),]  # unbuffered
                    + aw_args,
                    env=find_worker_env(idx),
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    close_fds=close_fds,
                )

            def run_logger(workR, widx):
                # this thread shuts itself down by reading from worker's stdout
                # which either reads content from stdout or blocks until it can do so
                # when the worker is shut down, empty byte string is returned continuously
                # which evaluates as None so the break is hit
                # I don't really like this approach; we should just shut it down with the other
                # threads explicitly
                while True:
                    try:
                        line = workR.stdout.readline()
                        line = line.decode("utf-8")
                        if not line:
                            break
                        log_msg_match = re.match(fr"{workR.pid}\|(10|20|30|40|50)\|(.*)", line)
                        if log_msg_match:
                            levelno = int(log_msg_match.group(1))
                            msg = log_msg_match.group(2)
                        else:
                            levelno = 20
                            msg = line

                        LOGGER.log(levelno, "\n\r  [Worker %d (%d)] %s", widx, workR.pid, msg.rstrip())

                    except Exception as e:
                        LOGGER.exception(e)
                        break

            start_daemon_thread(
                target=run_logger, args=(worker, idx), name="worker stdout logger thread"
            )

            self.workers += [worker]

    def stop_workers(self):
        if self.boundary is not None:
            LOGGER.debug("joining boundary")
            self.boundary.join()
            LOGGER.debug("destroying boundary zmq context")
            self.boundary.zmq_context.destroy(0)
            self.boundary = None
        for worker in self.workers:
            LOGGER.debug("waiting on worker")
            worker.wait()
        self.workers = []



from ._change_setting_event import ChangeSettingEvent


class AfterChangeSettingEvent(ChangeSettingEvent):
    """Indicates that a setting has changed its value

    """

    def __init__(self, old_value, new_value):
        ChangeSettingEvent.__init__(self, old_value, new_value)



from ._name import Name


class GridName(Name):
    """
    A setting that provides a GridInfo object
    """

    def __init__(self, text, value="Grid", *args, **kwargs):
        super(GridName, self).__init__(text, "gridgroup", value, *args, **kwargs)



# ImageFile must be initialized before ImagePlane
from ._image_file import ImageFile
from ._image_plane import ImagePlane
from ._image_set_channel_descriptor import ImageSetChannelDescriptor
from ._listener import Listener
from ._pipeline import Pipeline
from .dependency import Dependency
from .dependency import ImageDependency
from .dependency import MeasurementDependency
from .dependency import ObjectDependency
from .event import CancelledException
from .event import EndRun
from .event import Event
from .event import FileWalkEnded
from .event import FileWalkStarted
from .event import IPDLoadException
from .event import LoadException
from .event import ModuleAdded
from .event import ModuleDisabled
from .event import ModuleEdited
from .event import ModuleEnabled
from .event import ModuleMoved
from .event import ModuleRemoved
from .event import ModuleShowWindow
from .event import PipelineCleared
from .event import PipelineLoadCancelledException
from .event import PipelineLoaded
from .event import PrepareRunError
from .event import URLsAdded
from .event import URLsCleared
from .event import URLsRemoved
from .event.run_exception import PostRunException
from .event.run_exception import PrepareRunException
from .event.run_exception import RunException
from .io import dump
